# coding: utf-8
import abc
import array
import collections.abc
import os
import sys
import types
import typing

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal  # type: ignore

BUFFER = typing.Union[bytes, bytearray, memoryview]
IDENTITY_FILTER_PREFERENCE = Literal["conscover", "origorder", "random"]
COMPUTE_WEIGHTS_METHOD = Literal["pb", "gsc", "blosum"]
MSA_FORMAT = Literal[
    "stockholm", "pfam", "a2m", "psiblast", "selex", "afa", "clustal", 
    "clustallike", "phylip", "phylips"
]

# --- Alphabet ---------------------------------------------------------------

class Alphabet(object):
    @classmethod
    def dna(self) -> Alphabet: ...
    @classmethod
    def rna(self) -> Alphabet: ...
    @classmethod
    def amino(self) -> Alphabet: ...
    def __repr__(self) -> str: ...
    def __eq__(self, other: object) -> bool: ...
    def __reduce__(
        self,
    ) -> typing.Tuple[typing.Callable[[], Alphabet], typing.Tuple[()]]: ...
    def __sizeof__(self) -> int: ...
    @property
    def K(self) -> int: ...
    @property
    def Kp(self) -> int: ...
    @property
    def symbols(self) -> str: ...
    def is_dna(self) -> bool: ...
    def is_rna(self) -> bool: ...
    def is_amino(self) -> bool: ...
    def is_nucleotide(self) -> bool: ...
    def encode(self, sequence: str) -> VectorU8: ...
    def decode(
        self, sequence: typing.Union[VectorU8, memoryview, bytes, bytearray]
    ) -> str: ...

# --- GeneticCode ------------------------------------------------------------

class GeneticCode(object):
    def __init__(
        self,
        translation_table: int = 1,
        *,
        nucleotide_alphabet: Alphabet = ...,
        amino_alphabet: Alphabet = ...,
    ) -> None: ...
    @property
    def translation_table(self) -> int: ...
    @translation_table.setter
    def translation_table(self, translation_table: int) -> None: ...
    @property
    def description(self) -> str: ...
    def translate(
        self, sequence: typing.Union[VectorU8, memoryview, bytes, bytearray]
    ) -> VectorU8: ...

# --- Bitfield ---------------------------------------------------------------

class Bitfield(typing.Sequence[bool]):
    @classmethod
    def zeros(cls, n: int, /) -> Bitfield: ...
    @classmethod
    def ones(cls, n: int, /) -> Bitfield: ...
    def __init__(self, iterable: typing.Iterable[object]) -> None: ...
    def __len__(self) -> int: ...
    @typing.overload
    def __getitem__(self, index: int) -> bool: ...
    @typing.overload
    def __getitem__(self, index: slice) -> typing.Sequence[bool]: ...
    def __setitem__(self, index: int, value: bool) -> None: ...
    def __eq__(self, other: object) -> bool: ...
    def __getstate__(self) -> typing.Dict[str, object]: ...
    def __setstate__(self, state: typing.Dict[str, object]) -> None: ...
    def __sizeof__(self) -> int: ...
    def __copy__(self) -> Bitfield: ...
    def __deepcopy__(self, memo: typing.Dict[int, object]) -> Bitfield: ...
    def __buffer__(self, flags: int) -> memoryview: ...
    def count(self, value: bool = True) -> int: ...
    def copy(self) -> Bitfield: ...
    def toggle(self, index: int) -> None: ...

# --- KeyHash ----------------------------------------------------------------

class KeyHash(typing.Mapping[bytes, int]):
    def __init__(self) -> None: ...
    def __copy__(self) -> KeyHash: ...
    def __len__(self) -> int: ...
    def __contains__(self, value: object) -> bool: ...
    def __getitem__(self, item: bytes) -> int: ...
    def __iter__(self) -> typing.Iterator[bytes]: ...
    def __getstate__(self) -> typing.Dict[str, object]: ...
    def __setstate__(self, state: typing.Dict[str, object]) -> None: ...
    def __sizeof__(self) -> int: ...
    def clear(self) -> None: ...
    def copy(self) -> KeyHash: ...

# --- Matrix & Vector --------------------------------------------------------

_T = typing.TypeVar("_T")
_V = typing.TypeVar("_V")
_M = typing.TypeVar("_M")

class Vector(typing.Sequence[_T], typing.Generic[_T], abc.ABC):
    def __reduce__(
        self: _V,
    ) -> typing.Tuple[typing.Type[_V], typing.Tuple[array.array[typing.Any]]]: ...
    @abc.abstractmethod
    @classmethod
    def zeros(cls: typing.Type[_V], n: int) -> _V: ...
    @abc.abstractmethod
    def __init__(self, iterable: typing.Iterable[_T] = ()): ...
    def __copy__(self: _V) -> _V: ...
    @abc.abstractmethod
    def __eq__(self, other: object) -> bool: ...
    @typing.overload
    @abc.abstractmethod
    def __getitem__(self, index: int) -> _T: ...
    @typing.overload
    @abc.abstractmethod
    def __getitem__(self, index: slice) -> Vector[_T]: ...
    @typing.overload
    @abc.abstractmethod
    def __setitem__(self, index: int, value: _T) -> None: ...
    @typing.overload
    @abc.abstractmethod
    def __setitem__(self, index: slice, value: _T) -> None: ...
    @abc.abstractmethod
    def __iadd__(self: _V, other: typing.Union[_T, _V]) -> _V: ...
    @abc.abstractmethod
    def __imul__(self: _V, other: typing.Union[_T, _V]) -> _V: ...
    @abc.abstractmethod
    def __matmul__(self: _V, other: _V) -> _T: ...
    def __sizeof__(self) -> int: ...
    def __repr__(self) -> str: ...
    def __len__(self) -> int: ...
    def __buffer__(self, flags: int) -> memoryview: ...
    @property
    def shape(self) -> typing.Tuple[int]: ...
    @property
    def strides(self) -> typing.Tuple[int]: ...
    @property
    @abc.abstractmethod
    def itemsize(self) -> int: ...
    @property
    @abc.abstractmethod
    def format(self) -> str: ...
    @abc.abstractmethod
    def argmax(self) -> int: ...
    @abc.abstractmethod
    def argmin(self) -> int: ...
    @abc.abstractmethod
    def copy(self: _V) -> _V: ...
    @abc.abstractmethod
    def max(self) -> _T: ...
    @abc.abstractmethod
    def min(self) -> _T: ...
    @abc.abstractmethod
    def reverse(self) -> None: ...
    @abc.abstractmethod
    def sum(self) -> _T: ...

class VectorD(Vector[float]):
    @classmethod
    def zeros(cls, n: int) -> VectorD: ...
    def __init__(self, iterable: typing.Iterable[float] = ()): ...
    def __copy__(self) -> VectorD: ...
    def __eq__(self, other: object) -> bool: ...
    @typing.overload
    def __getitem__(self, index: int) -> float: ...
    @typing.overload
    def __getitem__(self, index: slice) -> VectorD: ...
    @typing.overload
    def __setitem__(self, index: int, value: float) -> None: ...
    @typing.overload
    def __setitem__(self, index: slice, value: float) -> None: ...
    def __neg__(self) -> VectorD: ...
    def __abs__(self) -> VectorD: ...
    def __add__(self, other: typing.Union[float, VectorD]) -> VectorD: ...
    def __iadd__(self, other: typing.Union[float, VectorD]) -> VectorD: ...
    def __sub__(self, other: typing.Union[float, VectorD]) -> VectorD: ...
    def __isub__(self, other: typing.Union[float, VectorD]) -> VectorD: ...
    def __mul__(self, other: typing.Union[float, VectorD]) -> VectorD: ...
    def __imul__(self, other: typing.Union[float, VectorD]) -> VectorD: ...
    def __truediv__(self, other: typing.Union[float, VectorD]) -> VectorD: ...
    def __itruediv__(self, other: typing.Union[float, VectorD]) -> VectorD: ...
    def __matmul__(self, other: VectorD) -> float: ...
    def argmax(self) -> int: ...
    def argmin(self) -> int: ...
    def copy(self) -> VectorD: ...
    def entropy(self) -> float: ...
    def max(self) -> float: ...
    def min(self) -> float: ...
    def normalize(self) -> None: ...
    def relative_entropy(self, other: VectorD) -> float: ...
    def reverse(self) -> None: ...
    def sum(self) -> float: ...
    @property
    def itemsize(self) -> int: ...
    @property
    def format(self) -> str: ...

class VectorF(Vector[float]):
    @classmethod
    def zeros(cls, n: int) -> VectorF: ...
    def __init__(self, iterable: typing.Iterable[float] = ()): ...
    def __copy__(self) -> VectorF: ...
    def __eq__(self, other: object) -> bool: ...
    @typing.overload
    def __getitem__(self, index: int) -> float: ...
    @typing.overload
    def __getitem__(self, index: slice) -> VectorF: ...
    @typing.overload
    def __setitem__(self, index: int, value: float) -> None: ...
    @typing.overload
    def __setitem__(self, index: slice, value: float) -> None: ...
    def __neg__(self) -> VectorF: ...
    def __abs__(self) -> VectorF: ...
    def __add__(self, other: typing.Union[float, VectorF]) -> VectorF: ...
    def __iadd__(self, other: typing.Union[float, VectorF]) -> VectorF: ...
    def __sub__(self, other: typing.Union[float, VectorF]) -> VectorF: ...
    def __isub__(self, other: typing.Union[float, VectorF]) -> VectorF: ...
    def __mul__(self, other: typing.Union[float, VectorF]) -> VectorF: ...
    def __imul__(self, other: typing.Union[float, VectorF]) -> VectorF: ...
    def __truediv__(self, other: typing.Union[float, VectorF]) -> VectorF: ...
    def __itruediv__(self, other: typing.Union[float, VectorF]) -> VectorF: ...
    def __matmul__(self, other: VectorF) -> float: ...
    def argmax(self) -> int: ...
    def argmin(self) -> int: ...
    def copy(self) -> VectorF: ...
    def entropy(self) -> float: ...
    def max(self) -> float: ...
    def min(self) -> float: ...
    def normalize(self) -> None: ...
    def relative_entropy(self, other: VectorF) -> float: ...
    def reverse(self) -> None: ...
    def sum(self) -> float: ...
    @property
    def itemsize(self) -> int: ...
    @property
    def format(self) -> str: ...

class VectorU8(Vector[int]):
    @classmethod
    def zeros(cls, n: int) -> VectorU8: ...
    def __init__(self, iterable: typing.Iterable[int] = ()): ...
    def __copy__(self) -> VectorU8: ...
    def __eq__(self, other: object) -> bool: ...
    @typing.overload
    def __getitem__(self, index: int) -> int: ...
    @typing.overload
    def __getitem__(self, index: slice) -> VectorU8: ...
    @typing.overload
    def __setitem__(self, index: int, value: int) -> None: ...
    @typing.overload
    def __setitem__(self, index: slice, value: int) -> None: ...
    def __add__(self, other: typing.Union[int, VectorU8]) -> VectorU8: ...
    def __iadd__(self, other: typing.Union[int, VectorU8]) -> VectorU8: ...
    def __sub__(self, other: typing.Union[int, VectorU8]) -> VectorU8: ...
    def __isub__(self, other: typing.Union[int, VectorU8]) -> VectorU8: ...
    def __mul__(self, other: typing.Union[int, VectorU8]) -> VectorU8: ...
    def __imul__(self, other: typing.Union[int, VectorU8]) -> VectorU8: ...
    def __floordiv__(self, other: typing.Union[int, VectorU8]) -> VectorU8: ...
    def __ifloordiv__(self, other: typing.Union[int, VectorU8]) -> VectorU8: ...
    def __matmul__(self, other: VectorU8) -> int: ...
    def argmax(self) -> int: ...
    def argmin(self) -> int: ...
    def copy(self) -> VectorU8: ...
    def max(self) -> int: ...
    def min(self) -> int: ...
    def reverse(self) -> None: ...
    def sum(self) -> int: ...
    @property
    def itemsize(self) -> int: ...
    @property
    def format(self) -> str: ...

class Matrix(typing.Sequence[Vector[_T]], typing.Generic[_T], abc.ABC):
    def __reduce__(
        self: _V,
    ) -> typing.Tuple[typing.Type[_V], typing.Tuple[typing.List[Vector[_T]]]]: ...
    @abc.abstractmethod
    @classmethod
    def zeros(cls: typing.Type[_M], m: int, n: int) -> _M: ...
    @abc.abstractmethod
    def __init__(self, iterable: typing.Iterable[typing.Iterable[_T]] = ()): ...
    @abc.abstractmethod
    def __copy__(self: _M) -> _M: ...
    def __len__(self) -> int: ...
    @abc.abstractmethod
    @typing.overload
    def __getitem__(self, index: int) -> Vector[_T]: ...
    @abc.abstractmethod
    @typing.overload
    def __getitem__(self, index: slice) -> Matrix[_T]: ...
    @abc.abstractmethod
    @typing.overload
    def __getitem__(self, index: typing.Tuple[int, int]) -> _T: ...
    @abc.abstractmethod
    def __setitem__(self, index: typing.Tuple[int, int], value: _T) -> None: ...
    @abc.abstractmethod
    def __add__(self: _M, other: typing.Union[_T, _M]) -> _M: ...
    @abc.abstractmethod
    def __iadd__(self: _M, other: typing.Union[_T, _M]) -> _M: ...
    @abc.abstractmethod
    def __mul__(self: _M, other: typing.Union[_T, _M]) -> _M: ...
    @abc.abstractmethod
    def __imul__(self: _M, other: typing.Union[_T, _M]) -> _M: ...
    @abc.abstractmethod
    def __repr__(self) -> str: ...
    def __buffer__(self, flags: int) -> memoryview: ...
    @property
    def shape(self) -> typing.Tuple[int, int]: ...
    @property
    def strides(self) -> typing.Tuple[int, int]: ...
    @property
    @abc.abstractmethod
    def itemsize(self) -> int: ...
    @property
    @abc.abstractmethod
    def format(self) -> str: ...
    @abc.abstractmethod
    def argmax(self) -> typing.Tuple[int, int]: ...
    @abc.abstractmethod
    def argmin(self) -> typing.Tuple[int, int]: ...
    @abc.abstractmethod
    def copy(self: _M) -> _M: ...
    @abc.abstractmethod
    def max(self) -> _T: ...
    @abc.abstractmethod
    def min(self) -> _T: ...
    @abc.abstractmethod
    def sum(self) -> _T: ...

class MatrixD(Matrix[float]):
    @classmethod
    def zeros(cls, m: int, n: int) -> MatrixD: ...
    def __init__(self, iterable: typing.Iterable[typing.Iterable[float]] = ()): ...
    def __copy__(self) -> MatrixD: ...
    def __eq__(self, other: object) -> bool: ...
    @typing.overload
    def __getitem__(self, index: int) -> VectorD: ...
    @typing.overload
    def __getitem__(self, index: slice) -> MatrixD: ...
    @typing.overload
    def __getitem__(self, index: typing.Tuple[int, int]) -> float: ...
    def __setitem__(self, index: typing.Tuple[int, int], value: float) -> None: ...
    def __abs__(self) -> MatrixD: ...
    def __add__(self, other: typing.Union[float, MatrixD]) -> MatrixD: ...
    def __iadd__(self, other: typing.Union[float, MatrixD]) -> MatrixD: ...
    def __mul__(self, other: typing.Union[float, MatrixD]) -> MatrixD: ...
    def __imul__(self, other: typing.Union[float, MatrixD]) -> MatrixD: ...
    def __repr__(self) -> str: ...
    def __sizeof__(self) -> int: ...
    def argmax(self) -> typing.Tuple[int, int]: ...
    def argmin(self) -> typing.Tuple[int, int]: ...
    def copy(self) -> MatrixD: ...
    def max(self) -> float: ...
    def min(self) -> float: ...
    def sum(self) -> float: ...
    @property
    def itemsize(self) -> int: ...
    @property
    def format(self) -> str: ...

class MatrixF(Matrix[float]):
    @classmethod
    def zeros(cls, m: int, n: int) -> MatrixF: ...
    def __init__(self, iterable: typing.Iterable[typing.Iterable[float]] = ()): ...
    def __copy__(self) -> MatrixF: ...
    def __eq__(self, other: object) -> bool: ...
    @typing.overload
    def __getitem__(self, index: int) -> VectorF: ...
    @typing.overload
    def __getitem__(self, index: slice) -> MatrixF: ...
    @typing.overload
    def __getitem__(self, index: typing.Tuple[int, int]) -> float: ...
    def __setitem__(self, index: typing.Tuple[int, int], value: float) -> None: ...
    def __abs__(self) -> MatrixF: ...
    def __add__(self, other: typing.Union[float, MatrixF]) -> MatrixF: ...
    def __iadd__(self, other: typing.Union[float, MatrixF]) -> MatrixF: ...
    def __mul__(self, other: typing.Union[float, MatrixF]) -> MatrixF: ...
    def __imul__(self, other: typing.Union[float, MatrixF]) -> MatrixF: ...
    def __repr__(self) -> str: ...
    def __sizeof__(self) -> int: ...
    def argmax(self) -> typing.Tuple[int, int]: ...
    def argmin(self) -> typing.Tuple[int, int]: ...
    def copy(self) -> MatrixF: ...
    def max(self) -> float: ...
    def min(self) -> float: ...
    def sum(self) -> float: ...
    @property
    def itemsize(self) -> int: ...
    @property
    def format(self) -> str: ...

class MatrixU8(Matrix[int]):
    @classmethod
    def zeros(cls, m: int, n: int) -> MatrixU8: ...
    def __init__(self, iterable: typing.Iterable[typing.Iterable[int]] = ()): ...
    def __copy__(self) -> MatrixU8: ...
    def __eq__(self, other: object) -> bool: ...
    @typing.overload
    def __getitem__(self, index: int) -> VectorU8: ...
    @typing.overload
    def __getitem__(self, index: slice) -> MatrixU8: ...
    @typing.overload
    def __getitem__(self, index: typing.Tuple[int, int]) -> int: ...
    def __setitem__(self, index: typing.Tuple[int, int], value: int) -> None: ...
    def __add__(self, other: typing.Union[int, MatrixU8]) -> MatrixU8: ...
    def __iadd__(self, other: typing.Union[int, MatrixU8]) -> MatrixU8: ...
    def __mul__(self, other: typing.Union[int, MatrixU8]) -> MatrixU8: ...
    def __imul__(self, other: typing.Union[int, MatrixU8]) -> MatrixU8: ...
    def __repr__(self) -> str: ...
    def __sizeof__(self) -> int: ...
    def argmax(self) -> typing.Tuple[int, int]: ...
    def argmin(self) -> typing.Tuple[int, int]: ...
    def copy(self) -> MatrixU8: ...
    def max(self) -> int: ...
    def min(self) -> int: ...
    def sum(self) -> int: ...
    @property
    def itemsize(self) -> int: ...
    @property
    def format(self) -> str: ...

# --- Multiple Sequences Alignment -------------------------------------------

class MSA(abc.ABC, typing.Sized):
    @abc.abstractmethod
    def __init__(
        self, nsequences: int, length: typing.Optional[int] = None
    ) -> None: ...
    def __copy__(self: _M) -> _M: ...
    def __eq__(self, other: object) -> bool: ...
    def __len__(self) -> int: ...
    @property
    def accession(self) -> typing.Optional[bytes]: ...
    @accession.setter
    def accession(self, accession: typing.Optional[bytes]) -> None: ...
    @property
    def author(self) -> typing.Optional[bytes]: ...
    @author.setter
    def author(self, author: typing.Optional[bytes]) -> None: ...
    @property
    def name(self) -> typing.Optional[bytes]: ...
    @name.setter
    def name(self, name: typing.Optional[bytes]) -> None: ...
    @property
    def description(self) -> typing.Optional[bytes]: ...
    @description.setter
    def description(self, description: typing.Optional[bytes]) -> None: ...
    @property
    def names(self) -> typing.Tuple[bytes]: ...
    @property
    def reference(self) -> typing.Optional[bytes]: ...
    @reference.setter
    def reference(self, reference: typing.Optional[bytes]) -> None: ...
    @property
    def model_mask(self) -> typing.Optional[bytes]: ...
    @model_mask.setter
    def model_mask(self, model_mask: typing.Optional[bytes]) -> None: ...
    @property
    def secondary_structure(self) -> typing.Optional[bytes]: ...
    @secondary_structure.setter
    def secondary_structure(self, secondary_structure: typing.Optional[bytes]) -> None: ...
    @property
    def surface_accessibility(self) -> typing.Optional[bytes]: ...
    @surface_accessibility.setter
    def surface_accessibility(self, surface_accessibility: typing.Optional[bytes]) -> None: ...
    @property
    def posterior_probabilities(self) -> typing.Optional[bytes]: ...
    @posterior_probabilities.setter
    def posterior_probabilities(self, posterior_probabilities: typing.Optional[bytes]) -> None: ...
    @property
    def sequence_weights(self) -> VectorD: ...
    @sequence_weights.setter
    def sequence_weights(self, sequence_weights: VectorD) -> None: ...
    @sequence_weights.deleter
    def sequence_weights(self) -> None: ...
    @property
    def indexed(self) -> typing.Mapping[bytes, Sequence]: ...
    @abc.abstractmethod
    def copy(self: _M) -> _M: ...
    def checksum(self) -> int: ...
    def compute_weights(
        self, 
        method: COMPUTE_WEIGHTS_METHOD = "pb", 
        max_identity: float = 0.62,
    ) -> VectorD: ...
    def mark_fragments(self, threshold: float) -> Bitfield: ...
    def select(self: _M, sequences: typing.Optional[typing.Iterable[int]] = None, columns: typing.Optional[typing.Iterable[int]] = None) -> _M: ...
    def write(self, fh: typing.BinaryIO, format: MSA_FORMAT) -> None: ...

class TextMSA(MSA):
    def __init__(
        self,
        *,
        name: typing.Optional[bytes] = None,
        description: typing.Optional[bytes] = None,
        accession: typing.Optional[bytes] = None,
        sequences: typing.Optional[typing.Iterable[TextSequence]] = None,
        author: typing.Optional[bytes] = None,
    ) -> None: ...
    def __copy__(self) -> TextMSA: ...
    def copy(self) -> TextMSA: ...
    def digitize(self, alphabet: Alphabet) -> DigitalMSA: ...
    @property
    def alignment(self) -> typing.Sequence[str]: ...
    @property
    def sequences(self) -> typing.Sequence[TextSequence]: ...
    @property
    def indexed(self) -> typing.Mapping[bytes, TextSequence]: ...

class DigitalMSA(MSA):
    alphabet: Alphabet
    @classmethod
    def sample(
        cls,
        alphabet: Alphabet,
        max_sequences: int,
        max_length: int,
        randomness: typing.Union[Randomness, int, None] = None,
    ) -> DigitalMSA: ...
    def __init__(
        self,
        *,
        alphabet: Alphabet,
        name: typing.Optional[bytes] = None,
        description: typing.Optional[bytes] = None,
        accession: typing.Optional[bytes] = None,
        sequences: typing.Optional[typing.Iterable[DigitalSequence]] = None,
        author: typing.Optional[bytes] = None,
    ) -> None: ...
    def __copy__(self) -> DigitalMSA: ...
    def copy(self) -> DigitalMSA: ...
    def textize(self) -> TextMSA: ...
    @property
    def alignment(self) -> typing.Sequence[VectorU8]: ...
    @property
    def sequences(self) -> typing.Sequence[DigitalSequence]: ...
    @property
    def indexed(self) -> typing.Mapping[bytes, DigitalSequence]: ...
    def compute_weights(
        self, 
        method: COMPUTE_WEIGHTS_METHOD = "pb", 
        max_identity: float = 0.62,
        *,
        fragment_threshold: float = 0.5,
        consensus_fraction: float = 0.5,
        ignore_rf: bool = False,
        sample: bool = True,
        sample_threshold: int = 50000,
        sample_count: int = 10000,
        max_fragments: int = 5000,
        seed: int = 42,
        preference: IDENTITY_FILTER_PREFERENCE = "conscover",
    ) -> VectorD: ...
    def identity_filter(
        self,
        max_identity: float = 0.8,
        *,
        fragment_threshold: float = 0.5,
        consensus_fraction: float = 0.5,
        ignore_rf: bool = False,
        sample: bool = True,
        sample_threshold: int = 50000,
        sample_count: int = 10000,
        max_fragments: int = 5000,
        seed: int = 42,
        preference: IDENTITY_FILTER_PREFERENCE = "conscover",
    ) -> DigitalMSA: ...
    @typing.overload
    def reverse_complement(self, inplace: Literal[True]) -> None: ...
    @typing.overload
    def reverse_complement(self, inplace: Literal[False] = False) -> DigitalMSA: ...


# --- MSA File ---------------------------------------------------------------

M = typing.TypeVar("M", TextMSA, DigitalMSA)

class MSAFile(typing.Generic[M], typing.ContextManager[MSAFile[M]], typing.Iterator[M]):
    _FORMATS: typing.ClassVar[typing.Dict[MSA_FORMAT, int]]
    alphabet: typing.Optional[Alphabet]
    name: typing.Optional[str]
    @typing.overload
    def __init__(
        self: MSAFile[DigitalMSA],
        file: typing.Union[typing.AnyStr, os.PathLike[typing.AnyStr], typing.BinaryIO],
        format: typing.Optional[MSA_FORMAT] = None,
        *,
        digital: Literal[True],
        alphabet: typing.Optional[Alphabet] = None,
    ) -> None: ...
    @typing.overload
    def __init__(
        self: MSAFile[TextMSA],
        file: typing.Union[typing.AnyStr, os.PathLike[typing.AnyStr], typing.BinaryIO],
        format: typing.Optional[str] = None,
        *,
        digital: Literal[False] = False,
        alphabet: typing.Optional[Alphabet] = None,
    ) -> None: ...
    def __enter__(self) -> MSAFile[M]: ...
    def __exit__(
        self,
        exc_type: typing.Optional[typing.Type[BaseException]],
        exc_value: typing.Optional[BaseException],
        traceback: typing.Optional[types.TracebackType],
    ) -> bool: ...
    def __iter__(self) -> MSAFile[M]: ...
    def __next__(self) -> M: ...
    def __repr__(self) -> str: ...
    @property
    def closed(self) -> bool: ...
    @property
    def digital(self) -> bool: ...
    @property
    def format(self) -> MSA_FORMAT: ...
    def read(self) -> typing.Optional[M]: ...
    def close(self) -> None: ...

# --- Randomness -------------------------------------------------------------

class Randomness(object):
    def __init__(
        self, seed: typing.Optional[int] = None, fast: bool = False
    ) -> None: ...
    def __copy__(self) -> Randomness: ...
    def __getstate__(self) -> typing.Tuple[typing.Any, ...]: ...
    def __setstate__(self, state: typing.Tuple[typing.Any, ...]) -> None: ...
    def __sizeof__(self) -> int: ...
    def __repr__(self) -> str: ...
    @property
    def fast(self) -> bool: ...
    def getstate(self) -> typing.Tuple[typing.Any, ...]: ...
    def setstate(self, state: typing.Tuple[typing.Any, ...]) -> None: ...
    def seed(self, n: typing.Optional[int] = None) -> None: ...
    def copy(self) -> Randomness: ...
    def random(self) -> float: ...
    def normalvariate(self, mu: float, sigma: float) -> float: ...

# --- Sequence ---------------------------------------------------------------

class Sequence(typing.Sized, abc.ABC):
    @abc.abstractmethod
    def __init__(self) -> None: ...
    def __copy__(self) -> Sequence: ...
    def __len__(self) -> int: ...
    def __eq__(self, other: object) -> bool: ...
    @property
    def accession(self) -> bytes: ...
    @accession.setter
    def accession(self, accession: bytes) -> None: ...
    @property
    def description(self) -> bytes: ...
    @description.setter
    def description(self, description: bytes) -> None: ...
    @property
    def name(self) -> bytes: ...
    @name.setter
    def name(self, name: bytes) -> None: ...
    @property
    def source(self) -> bytes: ...
    @source.setter
    def source(self, src: bytes) -> None: ...
    @property
    def residue_markups(self) -> typing.Dict[bytes, bytes]: ...
    @residue_markups.setter
    def residue_markups(self, xr: typing.Dict[bytes, bytes]) -> None: ...
    def checksum(self) -> int: ...
    def clear(self) -> None: ...
    @abc.abstractmethod
    def copy(self) -> Sequence: ...
    def write(self, fh: typing.BinaryIO) -> None: ...
    @typing.overload
    def reverse_complement(self, inplace: Literal[True]) -> None: ...
    @typing.overload
    def reverse_complement(self, inplace: Literal[False] = False) -> Sequence: ...

class TextSequence(Sequence):
    @classmethod
    def sample(
        cls,
        max_length: int,
        randomness: typing.Union[Randomness, int, None] = None,
    ) -> TextSequence: ...
    def __init__(
        self,
        *,
        name: typing.Optional[bytes] = None,
        description: typing.Optional[bytes] = None,
        accession: typing.Optional[bytes] = None,
        sequence: typing.Optional[str] = None,
        source: typing.Optional[bytes] = None,
        residue_markups: typing.Optional[typing.Dict[bytes, BUFFER]] = None,
    ) -> None: ...
    def __reduce__(
        self,
    ) -> typing.Tuple[typing.Type[TextSequence], typing.Tuple[object, ...]]: ...
    def __buffer__(self, flags: int) -> memoryview: ...
    @property
    def sequence(self) -> str: ...
    def copy(self) -> TextSequence: ...
    def digitize(self, alphabet: Alphabet) -> DigitalSequence: ...
    @typing.overload
    def reverse_complement(self, inplace: Literal[True]) -> None: ...
    @typing.overload
    def reverse_complement(self, inplace: Literal[False] = False) -> TextSequence: ...

class DigitalSequence(Sequence):
    alphabet: Alphabet
    @classmethod
    def sample(
        cls,
        alphabet: Alphabet,
        max_length: int,
        randomness: typing.Union[Randomness, int, None] = None,
    ) -> DigitalSequence: ...
    def __init__(
        self,
        alphabet: Alphabet,
        *,
        name: typing.Optional[bytes] = None,
        description: typing.Optional[bytes] = None,
        accession: typing.Optional[bytes] = None,
        sequence: typing.Union[BUFFER, VectorU8, None] = None,
        source: typing.Optional[bytes] = None,
        residue_markups: typing.Optional[typing.Dict[bytes, BUFFER]] = None,
    ) -> None: ...
    def __reduce__(
        self,
    ) -> typing.Tuple[typing.Type[DigitalSequence], typing.Tuple[object, ...]]: ...
    def __buffer__(self, flags: int) -> memoryview: ...
    @property
    def sequence(self) -> VectorU8: ...
    def copy(self) -> DigitalSequence: ...
    def textize(self) -> TextSequence: ...
    def translate(self, genetic_code: GeneticCode = ...) -> DigitalSequence: ...
    @typing.overload
    def reverse_complement(self, inplace: Literal[True]) -> None: ...
    @typing.overload
    def reverse_complement(
        self, inplace: Literal[False] = False
    ) -> DigitalSequence: ...

# --- Sequence block ---------------------------------------------------------

S = typing.TypeVar("S", TextSequence, DigitalSequence)
B = typing.TypeVar("B", SequenceBlock[TextSequence], SequenceBlock[DigitalSequence])

class SequenceBlock(typing.Generic[S], typing.MutableSequence[S]):
    def __len__(self) -> int: ...
    @typing.overload
    def __getitem__(self, index: int) -> S: ...
    @typing.overload
    def __getitem__(self: B, index: slice) -> B: ...
    @typing.overload
    def __setitem__(self, index: int, sequence: S) -> None: ...
    @typing.overload
    def __setitem__(self, index: slice, sequence: typing.Iterable[S]) -> None: ...
    @typing.overload
    def __delitem__(self, index: int) -> None: ...
    @typing.overload
    def __delitem__(self, index: slice) -> None: ...
    def __repr__(self) -> str: ...
    def __contains__(self, item: object) -> bool: ...
    def __eq__(self, other: object) -> bool: ...
    def __copy__(self: B) -> B: ...
    @property
    def indexed(self) -> typing.Mapping[bytes, S]: ...
    def clear(self) -> None: ...
    def append(self, sequence: S) -> None: ...
    def extend(self, iterable: typing.Iterable[S]) -> None: ...
    def pop(self, index: int = -1) -> S: ...
    def insert(self, index: int, sequence: S) -> None: ...
    def index(self, sequence: S, start: int = 0, end: int = sys.maxsize) -> int: ...
    def remove(self, sequence: S) -> None: ...
    def largest(self) -> S: ...

class TextSequenceBlock(SequenceBlock[TextSequence]):
    def __init__(self, iterable: typing.Iterable[TextSequence] = ()) -> None: ...
    def digitize(self, alphabet: Alphabet) -> DigitalSequenceBlock: ...

class DigitalSequenceBlock(SequenceBlock[DigitalSequence]):
    alphabet: Alphabet
    def __init__(
        self, alphabet: Alphabet, iterable: typing.Iterable[DigitalSequence] = ()
    ) -> None: ...
    def textize(self) -> TextSequenceBlock: ...
    def translate(self, genetic_code: GeneticCode = ...) -> DigitalSequence: ...

# --- Sequence File ----------------------------------------------------------

class SequenceFile(typing.Generic[S], typing.ContextManager[SequenceFile[S]], typing.Iterator[S]):
    _FORMATS: typing.ClassVar[typing.Dict[str, int]]
    alphabet: typing.Optional[Alphabet]
    name: typing.Optional[str]
    @classmethod
    def parse(
        cls, buffer: BUFFER, format: str, *, alphabet: typing.Optional[Alphabet] = None
    ) -> Sequence: ...
    @classmethod
    def parseinto(cls, seq: Sequence, buffer: BUFFER, format: str) -> Sequence: ...
    @typing.overload
    def __init__(
        self: SequenceFile[DigitalSequence],
        file: typing.Union[typing.AnyStr, os.PathLike[typing.AnyStr], typing.BinaryIO],
        format: typing.Optional[str] = None,
        *,
        ignore_gaps: bool = False,
        digital: Literal[True],
        alphabet: typing.Optional[Alphabet] = None,
    ) -> None: ...
    @typing.overload
    def __init__(
        self: SequenceFile[TextSequence],
        file: typing.Union[typing.AnyStr, os.PathLike[typing.AnyStr], typing.BinaryIO],
        format: typing.Optional[str] = None,
        *,
        ignore_gaps: bool = False,
        digital: Literal[False] = False,
        alphabet: typing.Optional[Alphabet] = None,
    ) -> None: ...
    def __enter__(self) -> SequenceFile[S]: ...
    def __exit__(
        self,
        exc_type: typing.Optional[typing.Type[BaseException]],
        exc_value: typing.Optional[BaseException],
        traceback: typing.Optional[types.TracebackType],
    ) -> bool: ...
    def __iter__(self) -> SequenceFile[S]: ...
    def __next__(self) -> S: ...
    def __repr__(self) -> str: ...
    @property
    def closed(self) -> bool: ...
    @property
    def digital(self) -> bool: ...
    @property
    def format(self) -> str: ...
    def read(
        self, skip_info: bool = False, skip_sequence: bool = False
    ) -> typing.Optional[S]: ...
    def readinto(
        self, seq: Sequence, skip_info: bool = False, skip_sequence: bool = False
    ) -> typing.Optional[S]: ...
    @typing.overload
    def read_block(
        self: SequenceFile[TextSequence],
        sequences: typing.Optional[int] = None,
        residues: typing.Optional[int] = None,
    ) -> TextSequenceBlock: ...
    @typing.overload
    def read_block(
        self: SequenceFile[DigitalSequence],
        sequences: typing.Optional[int] = None,
        residues: typing.Optional[int] = None,
    ) -> DigitalSequenceBlock: ...
    @typing.overload
    def read_block(
        self,
        sequences: typing.Optional[int] = None,
        residues: typing.Optional[int] = None,
    ) -> SequenceBlock[S]: ...
    def rewind(self) -> None: ...
    def close(self) -> None: ...
    def guess_alphabet(self) -> typing.Optional[Alphabet]: ...

# --- Sequence/Subsequence Index ---------------------------------------------

class SSIReader(object):
    class Entry(typing.NamedTuple):
        fd: int
        record_offset: int
        data_offset: int
        record_length: int

    class FileInfo(typing.NamedTuple):
        name: str
        format: int
    def __init__(
        self, file: typing.Union[typing.AnyStr, os.PathLike[typing.AnyStr]]
    ) -> None: ...
    def __enter__(self) -> SSIReader: ...
    def __exit__(
        self,
        exc_type: typing.Optional[typing.Type[BaseException]],
        exc_value: typing.Optional[BaseException],
        traceback: typing.Optional[types.TracebackType],
    ) -> bool: ...
    def close(self) -> None: ...

class SSIWriter(object):
    def __init__(
        self, file: typing.Union[typing.AnyStr, os.PathLike[typing.AnyStr]]
    ) -> None: ...
    def __enter__(self) -> SSIWriter: ...
    def __exit__(
        self,
        exc_type: typing.Optional[typing.Type[BaseException]],
        exc_value: typing.Optional[BaseException],
        traceback: typing.Optional[types.TracebackType],
    ) -> bool: ...
    def add_alias(self, alias: bytes, key: bytes) -> None: ...
    def add_file(self, filename: str, format: int = 0) -> int: ...
    def add_key(
        self,
        key: bytes,
        fd: int,
        record_offset: int,
        data_offset: int = 0,
        record_length: int = 0,
    ) -> None: ...
    def close(self) -> None: ...



# coding: utf-8
# cython: language_level=3, linetrace=True

# --- C imports --------------------------------------------------------------

from libc.stdint cimport int64_t, uint8_t, uint16_t, uint32_t, uint64_t
from posix.types cimport off_t

cimport libeasel.sq
from libeasel.alphabet cimport ESL_ALPHABET
from libeasel.bitfield cimport ESL_BITFIELD
from libeasel.gencode cimport ESL_GENCODE
from libeasel.keyhash cimport ESL_KEYHASH
from libeasel.msa cimport ESL_MSA
from libeasel.msafile cimport ESL_MSAFILE
from libeasel.random cimport ESL_RANDOMNESS
from libeasel.sq cimport ESL_SQ, ESL_SQ_BLOCK
from libeasel.sqio cimport ESL_SQFILE
from libeasel.ssi cimport ESL_SSI, ESL_NEWSSI
from libeasel cimport ESL_DSQ


# --- Alphabet ---------------------------------------------------------------

cdef class Alphabet:
    cdef ESL_ALPHABET* _abc

    cdef int _init_default(self, int ty) except 1 nogil
    cdef inline bint _eq(self, Alphabet other) nogil

    cpdef bint is_dna(self)
    cpdef bint is_rna(self)
    cpdef bint is_amino(self)
    cpdef bint is_nucleotide(self)
    cpdef VectorU8 encode(self, str sequence)
    cpdef str decode(self, const ESL_DSQ[::1] sequence)


# --- GeneticCode ------------------------------------------------------------

cdef class GeneticCode:
    cdef readonly Alphabet     amino_alphabet
    cdef readonly Alphabet     nucleotide_alphabet
    cdef          ESL_GENCODE* _gcode

    cdef int _translate(
        self,
        const ESL_DSQ* seq,
        int64_t seqlen,
        ESL_DSQ* out,
        int64_t outlen
    ) except -1 nogil

    cpdef VectorU8 translate(self, const ESL_DSQ[::1] sequence)


# --- Bitfield ---------------------------------------------------------------

cdef class Bitfield:
    cdef ESL_BITFIELD* _b
    cdef readonly Py_ssize_t _shape[1]

    cdef size_t _wrap_index(self, int index) except -1
    cpdef size_t count(self, bint value=*)
    cpdef Bitfield copy(self)
    cpdef void toggle(self, int index) except *


# --- KeyHash ----------------------------------------------------------------

cdef class KeyHash:
    cdef ESL_KEYHASH* _kh

    cpdef int add(self, bytes key) except -1
    cpdef void clear(self) except *
    cpdef KeyHash copy(self)


# --- Matrix & Vector --------------------------------------------------------

cdef class Vector:
    cdef object              _owner
    cdef int                 _n
    cdef readonly Py_ssize_t _shape[1]
    cdef void*               _data

    cdef const char* _format(self) noexcept
    cdef int _allocate(self, size_t n) except -1

cdef class VectorD(Vector):
    cpdef int argmax(self) except -1
    cpdef int argmin(self) except -1
    cpdef VectorD copy(self)
    cpdef double entropy(self) except *
    cpdef double max(self) except *
    cpdef double min(self) except *
    cpdef object normalize(self)
    cpdef double relative_entropy(self, VectorD other) except *
    cpdef object reverse(self)
    cpdef double sum(self)

cdef class VectorF(Vector):
    cpdef int argmax(self) except -1
    cpdef int argmin(self) except -1
    cpdef VectorF copy(self)
    cpdef float entropy(self) except *
    cpdef float max(self) except *
    cpdef float min(self) except *
    cpdef object normalize(self)
    cpdef float relative_entropy(self, VectorF other) except *
    cpdef object reverse(self)
    cpdef float sum(self)

cdef class VectorU8(Vector):
    cpdef int argmax(self) except -1
    cpdef int argmin(self) except -1
    cpdef VectorU8 copy(self)
    cpdef uint8_t max(self) except *
    cpdef uint8_t min(self) except *
    cpdef object reverse(self)
    cpdef uint8_t sum(self)

cdef class Matrix:
    cdef object              _owner
    cdef int                 _n
    cdef int                 _m
    cdef readonly Py_ssize_t _shape[2]
    cdef void**              _data

    cdef const char* _format(self) noexcept
    cdef int _allocate(self, size_t m, size_t n) except -1

cdef class MatrixD(Matrix):
    cpdef tuple argmax(self)
    cpdef tuple argmin(self)
    cpdef MatrixD copy(self)
    cpdef double max(self)
    cpdef double min(self)
    cpdef double sum(self)

cdef class MatrixF(Matrix):
    cpdef tuple argmax(self)
    cpdef tuple argmin(self)
    cpdef MatrixF copy(self)
    cpdef float max(self)
    cpdef float min(self)
    cpdef float sum(self)

cdef class MatrixU8(Matrix):
    cpdef tuple argmax(self)
    cpdef tuple argmin(self)
    cpdef MatrixU8 copy(self)
    cpdef uint8_t max(self)
    cpdef uint8_t min(self)
    cpdef uint8_t sum(self)


# --- Multiple Sequences Alignment -------------------------------------------

cdef class MSA:
    cdef ESL_MSA* _msa

    cdef int _rehash(self) except 1 nogil
    cdef int _set_annotation(self, char** field, char* value) except 1 nogil

    cpdef uint32_t checksum(self)
    cpdef Bitfield mark_fragments(self, float threshold)
    cpdef MSA select(self, sequences = ?, columns = ?)
    cpdef VectorD compute_weights(self, str method = *, float max_identity = *)
    cpdef void write(self, object fh, str format) except *


cdef class TextMSA(MSA):
    cdef int _set_sequence(self, int idx, ESL_SQ* seq) except 1 nogil
    cpdef TextMSA copy(self)
    cpdef DigitalMSA digitize(self, Alphabet alphabet)


cdef class DigitalMSA(MSA):
    cdef readonly Alphabet alphabet

    cdef int _set_sequence(self, int idx, ESL_SQ* seq) except 1 nogil
    cpdef DigitalMSA copy(self)
    cpdef TextMSA textize(self)

    cpdef DigitalMSA reverse_complement(self, bint inplace=*)
    cpdef DigitalMSA identity_filter(
        self,
        float max_identity=*,
        float fragment_threshold=*,
        float consensus_fraction=*,
        bint ignore_rf=*,
        bint sample=*,
        int sample_threshold=*,
        int sample_count=*,
        int max_fragments=*,
        uint64_t seed=?,
        str preference=*,
    )
    cpdef VectorD compute_weights(
        self,
        str method = *,
        float max_identity=*,
        float fragment_threshold=*,
        float consensus_fraction=*,
        bint ignore_rf=*,
        bint sample=*,
        int sample_threshold=*,
        int sample_count=*,
        int max_fragments=*,
        uint64_t seed=?,
        str preference=*,
    )


# --- MSA File ---------------------------------------------------------------

cdef class MSAFile:
    cdef          ESL_MSAFILE* _msaf
    cdef readonly object       _file
    cdef readonly Alphabet     alphabet
    cdef readonly str          name

    @staticmethod
    cdef ESL_MSAFILE* _open_fileobj(object fh, int fmt) except NULL

    cpdef void close(self)
    cpdef Alphabet guess_alphabet(self)
    cpdef MSA read(self)


# --- Randomness -------------------------------------------------------------

cdef class Randomness:
    cdef ESL_RANDOMNESS* _rng
    cdef object          _owner

    cpdef void seed(self, object n=*) except *
    cpdef Randomness copy(self)
    cpdef double random(self)
    cpdef double normalvariate(self, double mu, double sigma)


ctypedef fused RandomnessOrSeed:
    Randomness
    object


# --- Sequence ---------------------------------------------------------------

cdef class Sequence:
    cdef ESL_SQ* _sq

    cpdef void clear(self) except *
    cpdef uint32_t checksum(self)
    cpdef void write(self, object fh) except *


cdef class TextSequence(Sequence):
    cpdef TextSequence copy(self)
    cpdef DigitalSequence digitize(self, Alphabet alphabet)
    cpdef TextSequence reverse_complement(self, bint inplace=*)


cdef class DigitalSequence(Sequence):
    cdef readonly Alphabet alphabet

    cpdef DigitalSequence copy(self)
    cpdef TextSequence textize(self)
    cpdef DigitalSequence translate(self, GeneticCode genetic_code=*)
    cpdef DigitalSequence reverse_complement(self, bint inplace=*)


# --- Sequence Block ---------------------------------------------------------

cdef class SequenceBlock:
    cdef          size_t     _length   # the number of sequences in the block
    cdef          size_t     _capacity # the total number of sequences that can be stored
    cdef          ESL_SQ**   _refs     # the array to pass the sequence references
    cdef          list       _storage  # the actual Python list where `Sequence` objects are stored
    cdef          object     _owner    # the owner, if the object is just a shallow copy
    cdef          ssize_t    _largest  # the index of the largest sequence, or -1
    cdef readonly KeyHash    _indexed  # a mapping of sequence names to sequences indices

    cdef void _on_modification(self) noexcept
    cdef int _rehash(self) except 1 nogil

    cdef void _allocate(self, size_t n) except *
    cdef void _append(self, Sequence sequence) except *
    cdef Sequence _pop(self, ssize_t index=*)
    cdef void _insert(self, ssize_t index, Sequence sequence) except *
    cdef size_t _index(self, Sequence sequence, ssize_t start=*, ssize_t stop=*) except *
    cdef void _remove(self, Sequence sequence) except *

    cpdef void extend(self, object iterable) except *
    cpdef SequenceBlock copy(self)
    cpdef void clear(self) except *
    cpdef Sequence largest(self)


cdef class TextSequenceBlock(SequenceBlock):
    cpdef void append(self, TextSequence sequence) except *
    cpdef TextSequence pop(self, ssize_t index=*)
    cpdef void insert(self, ssize_t index, TextSequence sequence) except *
    cpdef size_t index(self, TextSequence sequence, ssize_t start=*, ssize_t stop=*) except *
    cpdef void remove(self, TextSequence sequence) except *

    cpdef TextSequenceBlock copy(self)
    cpdef DigitalSequenceBlock digitize(self, Alphabet alphabet)
    cpdef TextSequence largest(self)

cdef class DigitalSequenceBlock(SequenceBlock):
    cdef readonly Alphabet   alphabet

    cpdef void append(self, DigitalSequence sequence) except *
    cpdef DigitalSequence pop(self, ssize_t index=*)
    cpdef void insert(self, ssize_t index, DigitalSequence sequence) except *
    cpdef size_t index(self, DigitalSequence sequence, ssize_t start=*, ssize_t stop=*) except *
    cpdef void remove(self, DigitalSequence sequence) except *

    cpdef DigitalSequenceBlock copy(self)
    cpdef TextSequenceBlock textize(self)
    cpdef DigitalSequenceBlock translate(self, GeneticCode genetic_code=*)
    cpdef DigitalSequence largest(self)

# --- Sequence File ----------------------------------------------------------

cdef class SequenceFile:
    cdef          ESL_SQFILE* _sqfp
    cdef readonly object      _file
    cdef readonly Alphabet    alphabet
    cdef readonly str         name

    @staticmethod
    cdef ESL_SQFILE* _open_fileobj(object fh, int fmt) except NULL

    cpdef void close(self) except *
    cpdef Alphabet guess_alphabet(self)
    cpdef Sequence read(self, bint skip_info=*, bint skip_sequence=*)
    cpdef Sequence readinto(self, Sequence, bint skip_info=*, bint skip_sequence=*)
    cpdef SequenceBlock read_block(self, object sequences=*, object residues=*)
    cpdef void rewind(self) except *

# --- Sequence/Subsequence Index ---------------------------------------------

cdef class SSIReader:
    cdef ESL_SSI* _ssi

    cpdef void close(self)


cdef class SSIWriter:
    cdef ESL_NEWSSI* _newssi

    cdef void _on_write(self)

    cpdef void     add_alias(self, bytes alias, bytes key) except *
    cpdef uint16_t add_file(self, object filename, int format = *) except *
    cpdef void     add_key(
        self,
        bytes key,
        uint16_t fd,
        off_t record_offset,
        off_t data_offset = *,
        int64_t record_length = *
    ) except *
    cpdef void close(self) except *



from libc.stdio cimport FILE
from libc.stdint cimport int8_t

from libeasel cimport ESL_DSQ
from libeasel.alphabet cimport ESL_ALPHABET
from libeasel.sq cimport ESL_SQ, ESL_SQ_BLOCK


cdef extern from "esl_gencode.h" nogil:

    ctypedef struct ESL_GENCODE:
        int     transl_table
        char[128]  desc

        ESL_DSQ[64] basic
        int8_t[64]  is_initiator

        ESL_ALPHABET* nt_abc
        ESL_ALPHABET* aa_abc

    ctypedef esl_gencode_workstate_s ESL_GENCODE_WORKSTATE
    ctypedef struct esl_gencode_workstate_s:
        # stateful info (which may get updated with each new seq, strand, and/or window):
        ESL_SQ[3] *psq
        int8_t[3]  in_orf
        int     apos
        int     frame
        int     codon
        int     inval
        bint    is_revcomp
        int     orfcount

        ESL_SQ_BLOCK* orf_block

        # one-time configuration information (from options)
        bint     do_watson;
        bint     do_crick;
        bint     using_initiators
        int     minlen
        FILE   *outfp
        int     outformat


    # Create/Destroy workstate
    # extern void esl_gencode_WorkstateDestroy(ESL_GENCODE_WORKSTATE *wrk);
    # extern ESL_GENCODE_WORKSTATE * esl_gencode_WorkstateCreate(ESL_GETOPTS *go, ESL_GENCODE *gcode);

    # the ESL_GENCODE genetic code object
    ESL_GENCODE *esl_gencode_Create(const ESL_ALPHABET *nt_abc, const ESL_ALPHABET *aa_abc);
    void         esl_gencode_Destroy            (ESL_GENCODE *gcode);
    int          esl_gencode_Set                (ESL_GENCODE *gcode,  int ncbi_transl_table);
    int          esl_gencode_SetInitiatorAny    (ESL_GENCODE *gcode);
    int          esl_gencode_SetInitiatorOnlyAUG(ESL_GENCODE *gcode);

    # reading and writing genetic codes in NCBI format
    # int          esl_gencode_Read(ESL_FILEPARSER *efp, const ESL_ALPHABET *nucleic_abc, const ESL_ALPHABET *amino_abc, ESL_GENCODE **ret_gcode);
    # int          esl_gencode_Write(FILE *ofp, const ESL_GENCODE *gcode, int add_comment);

    # DNA->protein digital translation, allowing ambiguity chars
    int   esl_gencode_GetTranslation(const ESL_GENCODE *gcode, ESL_DSQ *dsqp);
    int   esl_gencode_IsInitiator   (const ESL_GENCODE *gcode, ESL_DSQ *dsqp);

    # Debugging/development utilities
    char *esl_gencode_DecodeDigicodon(const ESL_GENCODE *gcode, int digicodon, char *codon);
    int   esl_gencode_DumpAltCodeTable(FILE *ofp);
    int   esl_gencode_Compare(const ESL_GENCODE *gc1, const ESL_GENCODE *gc2, int metadata_too);

    # Functions for processing ORFs
    int esl_gencode_ProcessOrf(ESL_GENCODE_WORKSTATE *wrk, ESL_SQ *sq);
    void esl_gencode_ProcessStart(ESL_GENCODE *gcode, ESL_GENCODE_WORKSTATE *wrk, ESL_SQ *sq);
    int esl_gencode_ProcessPiece(ESL_GENCODE *gcode, ESL_GENCODE_WORKSTATE *wrk, ESL_SQ *sq);
    int esl_gencode_ProcessEnd(ESL_GENCODE_WORKSTATE *wrk, ESL_SQ *sq);



# coding: utf-8
# cython: language_level=3, linetrace=True

# --- C imports --------------------------------------------------------------

from libc.stdint cimport uint8_t, uint32_t
from posix.types cimport off_t
from cpython.pythread cimport PyThread_type_lock

from libeasel import eslINFINITY
from libeasel.sq cimport ESL_SQ
from libeasel.sqio cimport ESL_SQFILE
from libhmmer.p7_alidisplay cimport P7_ALIDISPLAY
from libhmmer.p7_bg cimport P7_BG
from libhmmer.p7_builder cimport P7_BUILDER
from libhmmer.p7_domain cimport P7_DOMAIN
from libhmmer.p7_hit cimport P7_HIT
from libhmmer.p7_hmm cimport P7_HMM, p7_NOFFSETS, p7_NEVPARAM, p7_NCUTOFFS
from libhmmer.p7_hmmfile cimport P7_HMMFILE
from libhmmer.p7_pipeline cimport P7_PIPELINE, p7_pipemodes_e, p7_zsetby_e
from libhmmer.p7_profile cimport P7_PROFILE
from libhmmer.p7_scoredata cimport P7_SCOREDATA
from libhmmer.p7_tophits cimport P7_TOPHITS
from libhmmer.p7_trace cimport P7_TRACE
from libhmmer.nhmmer cimport ID_LENGTH_LIST

if HMMER_IMPL == "VMX":
    from libhmmer.impl_vmx.p7_omx cimport P7_OM_BLOCK
    from libhmmer.impl_vmx.p7_oprofile cimport P7_OPROFILE
elif HMMER_IMPL == "SSE":
    from libhmmer.impl_sse.p7_omx cimport P7_OM_BLOCK
    from libhmmer.impl_sse.p7_oprofile cimport P7_OPROFILE
elif HMMER_IMPL == "NEON":
    from libhmmer.impl_neon.p7_omx cimport P7_OM_BLOCK
    from libhmmer.impl_neon.p7_oprofile cimport P7_OPROFILE

from .easel cimport (
    Alphabet,
    DigitalSequence,
    DigitalSequenceBlock,
    DigitalMSA,
    KeyHash,
    MSA,
    Randomness,
    VectorF,
    VectorU8,
    SequenceFile,
)


# --- Fused types ------------------------------------------------------------

ctypedef fused ScanTargets:
    HMMPressedFile
    OptimizedProfileBlock

ctypedef fused SearchTargets:
    SequenceFile
    DigitalSequenceBlock

ctypedef fused SearchQuery:
    HMM
    Profile
    OptimizedProfile


# --- Cython classes ---------------------------------------------------------


cdef class Alignment:
    cdef readonly Domain domain
    cdef P7_ALIDISPLAY* _ad

cdef class Background:
    cdef readonly bint     uniform
    cdef readonly Alphabet alphabet
    cdef readonly VectorF  residue_frequencies
    cdef          P7_BG*   _bg
    cdef          int      _L

    cpdef Background copy(self)


cdef class Builder:
    cdef readonly object prior_scheme
    cdef readonly object effective_number
    cdef readonly object architecture
    cdef readonly object weighting
    cdef public   str    score_matrix
    cdef public   double popen
    cdef public   double pextend
    cdef readonly Alphabet alphabet
    cdef readonly Randomness randomness

    cdef uint32_t    _seed    # the seed passed at builder initialization
    cdef P7_BUILDER* _bld

    cpdef tuple build(self, DigitalSequence sequence, Background background)
    cpdef tuple build_msa(self, DigitalMSA msa, Background background)
    cpdef Builder copy(self)


cdef class Cutoffs:
    cdef object              _owner
    cdef int*                _flags
    cdef bint                _is_profile
    cdef float[p7_NCUTOFFS]* _cutoffs

    cpdef VectorF as_vector(self)
    cpdef bint gathering_available(self)
    cpdef bint trusted_available(self)
    cpdef bint noise_available(self)


cdef class Domain:
    cdef readonly Alignment alignment
    cdef readonly Hit hit
    cdef P7_DOMAIN* _dom

cdef class Domains:
    cdef readonly Hit hit


cdef class EvalueParameters:
    cdef object              _owner
    cdef float[p7_NEVPARAM]* _evparams

    cpdef VectorF as_vector(self)


cdef class Hit:
    # a reference to the TopHits that owns the wrapped P7_HIT, kept so that
    # the internal data is never deallocated before the Python class.
    cdef readonly TopHits hits
    cdef P7_HIT* _hit


cdef class HMM:
    cdef P7_HMM* _hmm
    # a reference to the Alphabet Python object to avoid deallocation of the
    # inner ESL_ALPHABET; the Python object provides reference counting for free
    cdef readonly Alphabet          alphabet

    cdef void _initialize(self) noexcept nogil
    cdef void _set_annotation_line(self, str line, char** ptr, int flag) except *

    cpdef HMM copy(self)
    cpdef VectorF match_occupancy(self)
    cpdef double mean_match_entropy(self) except *
    cpdef double mean_match_information(self, Background background) except *
    cpdef double mean_match_relative_entropy(self, Background background) except *
    cpdef void renormalize(self)
    cpdef void scale(self, double scale, bint exponential=?)
    cpdef void set_composition(self) except *
    cpdef void set_consensus(self, DigitalSequence sequence=?) except *
    cpdef Profile to_profile(self,  Background background=?,  int L=*,  bint multihit =*,  bint local=*)
    cpdef void validate(self, float tolerance=*) except *
    cpdef void write(self, object fh, bint binary=*) except *
    cpdef void zero(self) noexcept


cdef class HMMFile:
    cdef str         _name
    cdef P7_HMMFILE* _hfp
    cdef Alphabet    _alphabet
    cdef object      _file

    cpdef void close(self) except *
    cpdef void rewind(self) except *
    cpdef HMM read(self)
    cpdef bint is_pressed(self) except *
    cpdef HMMPressedFile optimized_profiles(self)

    @staticmethod
    cdef P7_HMMFILE* _open_fileobj(object fh) except *


cdef class HMMPressedFile:
    cdef P7_HMMFILE* _hfp
    cdef Alphabet    _alphabet
    cdef HMMFile     _hmmfile
    cdef size_t      _position

    cpdef void close(self) except *
    cpdef void rewind(self) except *
    cpdef OptimizedProfile read(self)


cdef class IterationResult:
    cdef readonly TopHits    hits
    cdef readonly DigitalMSA msa
    cdef readonly HMM        hmm
    cdef readonly bint       converged
    cdef readonly size_t     iteration


cdef class IterativeSearch:
    cdef readonly object               query
    cdef readonly Pipeline             pipeline
    cdef readonly Background           background
    cdef readonly Builder              builder
    cdef readonly bint                 converged
    cdef readonly DigitalSequenceBlock targets
    cdef readonly KeyHash              ranking
    cdef readonly size_t               iteration
    cdef          DigitalMSA           msa
    cdef          object               select_hits

    cpdef TopHits _search_hmm(self, HMM hmm)


cdef class OptimizedProfile:
    cdef P7_OPROFILE* _om
    cdef readonly Alphabet alphabet

    cpdef OptimizedProfile copy(self)
    cpdef void convert(self, Profile profile) except *
    cpdef void write(self, object fh_filter, object fh_profile) except *
    cpdef object ssv_filter(self, DigitalSequence seq)


cdef class OptimizedProfileBlock:
    cdef          PyThread_type_lock* _locks
    cdef          P7_OM_BLOCK*        _block
    cdef          list                _storage
    cdef readonly Alphabet            alphabet

    cdef void _allocate(self, size_t n) except *

    cpdef void append(self, OptimizedProfile optimized_profile) except *
    cpdef void clear(self) except *
    cpdef void extend(self, object iterable) except *
    cpdef size_t index(self, OptimizedProfile optimized_profile, ssize_t start=*, ssize_t stop=*) except *
    cpdef void insert(self, ssize_t index, OptimizedProfile optimized_profile) except *
    cpdef OptimizedProfile pop(self, ssize_t index=*)
    cpdef void remove(self, OptimizedProfile optimized_profile) except *
    cpdef OptimizedProfileBlock copy(self)


cdef class Offsets:
    cdef object              _owner
    cdef off_t[p7_NOFFSETS]* _offs


cdef class Pipeline:
    cdef          object     _Z            # either `Z` as an int, or `None`
    cdef          object     _domZ         # either `domZ` as an int, or `None`
    cdef          uint32_t   _seed         # the seed passed at pipeline initialization
    cdef          dict       _cutoff_save  # a local save of the reporting parameters

    cdef readonly Alphabet         alphabet
    cdef readonly Background       background
    cdef readonly Profile          profile
    cdef readonly OptimizedProfile opt
    cdef readonly Randomness       randomness

    cdef OptimizedProfile _optimized
    cdef P7_PIPELINE* _pli

    @staticmethod
    cdef int _missing_cutoffs(const P7_PIPELINE* pli, const P7_OPROFILE* om) except 1 nogil
    cdef int _save_cutoff_parameters(self) except 1
    cdef int _restore_cutoff_parameters(self) except 1
    cdef P7_OPROFILE* _get_om_from_query(self, SearchQuery query, int L = *) except NULL

    cpdef list    arguments(self)
    cpdef void    clear(self)

    cpdef TopHits search_hmm(
        self,
        SearchQuery query,
        SearchTargets sequences
    )
    cpdef TopHits search_msa(
        self,
        DigitalMSA query,
        SearchTargets sequences,
        Builder builder = ?
    )
    cpdef TopHits search_seq(
        self,
        DigitalSequence query,
        SearchTargets sequences,
        Builder builder = ?
    )
    @staticmethod
    cdef  int  _search_loop(
              P7_PIPELINE* pli,
              P7_OPROFILE* om,
              P7_BG*       bg,
        const ESL_SQ**     sq,
        const size_t       n_targets,
              P7_TOPHITS*  th,
    ) except 1 nogil
    @staticmethod
    cdef  int  _search_loop_file(
              P7_PIPELINE* pli,
              P7_OPROFILE* om,
              P7_BG*       bg,
              ESL_SQFILE*  sqfp,
              P7_TOPHITS*  th,
    ) except 1 nogil

    cpdef TopHits scan_seq(
        self,
        DigitalSequence query,
        ScanTargets targets
    )
    @staticmethod
    cdef int _scan_loop(
              P7_PIPELINE*        pli,
        const ESL_SQ*             sq,
              P7_BG*              bg,
              P7_OPROFILE**       om,
        const size_t              n_targets,
              P7_TOPHITS*         th,
              PyThread_type_lock* locks,
    ) except 1 nogil
    @staticmethod
    cdef int _scan_loop_file(
              P7_PIPELINE*     pli,
        const ESL_SQ*          sq,
              P7_BG*           bg,
              P7_HMMFILE*      hfp,
              P7_TOPHITS*      th,
    ) except 1 nogil

    cpdef IterativeSearch iterate_hmm(
        self,
        HMM query,
        DigitalSequenceBlock sequences,
        Builder builder = ?,
        object select_hits = ?,
    )
    cpdef IterativeSearch iterate_seq(
        self,
        DigitalSequence query,
        DigitalSequenceBlock sequences,
        Builder builder = ?,
        object select_hits = ?,
    )


cdef class LongTargetsPipeline(Pipeline):
    cdef int             _window_length
    cdef double          _window_beta
    cdef ID_LENGTH_LIST* _idlens

    cdef P7_OPROFILE* _get_om_from_query(self, SearchQuery query, int L = *) except NULL

    cpdef TopHits search_hmm(
        self,
        SearchQuery query,
        SearchTargets sequences
    )
    cpdef TopHits search_msa(
        self,
        DigitalMSA query,
        SearchTargets sequences,
        Builder builder = ?
    )
    cpdef TopHits search_seq(
        self,
        DigitalSequence query,
        SearchTargets sequences,
        Builder builder = ?
    )
    @staticmethod
    cdef int _search_loop_longtargets(
              P7_PIPELINE*    pli,
              P7_OPROFILE*    om,
              P7_BG*          bg,
        const ESL_SQ**        sq,
        const size_t          n_targets,
              P7_TOPHITS*     th,
              P7_SCOREDATA*   scoredata,
              ID_LENGTH_LIST* idlens
    ) except 1 nogil
    @staticmethod
    cdef int _search_loop_longtargets_file(
        P7_PIPELINE*  pli,
        P7_OPROFILE*  om,
        P7_BG*        bg,
        ESL_SQFILE*   sqfp,
        P7_TOPHITS*   th,
        P7_SCOREDATA* scoredata,
        ID_LENGTH_LIST* idlens
    ) except 1 nogil
    cpdef TopHits scan_seq(
        self,
        DigitalSequence query,
        ScanTargets targets
    )

cdef class Profile:
    cdef          P7_PROFILE* _gm
    cdef readonly Alphabet    alphabet

    cpdef void clear(self) except *
    cpdef void configure(self, HMM hmm, Background background, int L=?, bint multihit=*, bint local=*) except *
    cpdef Profile copy(self)
    cpdef OptimizedProfile to_optimized(self)


cdef class ScoreData:
    cdef P7_SCOREDATA* _sd
    cdef readonly int  Kp

    cpdef ScoreData copy(self)


cdef class TopHits:

    # NOTE(@althonos): this is not a full pipeline, but a local copy of the
    #                  accounting parameters so that the e-value can be
    #                  computed and thresholding can be done correctly.
    cdef P7_PIPELINE _pli
    cdef P7_TOPHITS* _th
    cdef object      _query
    cdef bint        _empty

    # cdef bytes       _qname
    # cdef bytes       _qacc
    # cdef int         _qlen

    cdef int _threshold(self, Pipeline pipeline) except 1 nogil
    cdef int _sort_by_key(self) except 1 nogil
    cdef int _sort_by_seqidx(self) except 1 nogil
    cdef void _check_threshold_parameters(self, const P7_PIPELINE* other) except *

    cpdef TopHits copy(self)
    cpdef int compare_ranking(self, KeyHash) except -1
    cpdef bint is_sorted(self, str by=*) except *
    cpdef void sort(self, str by=*) except *
    cpdef MSA to_msa(self, Alphabet alphabet, list sequences=?, list traces=?, bint trim=*, bint digitize=?, bint all_consensus_cols=?)
    cpdef void write(self, object fh, str format=*, bint header=*) except *


cdef class Trace:
    cdef P7_TRACE* _tr

    @staticmethod
    cdef bint _eq(P7_TRACE* tr1, P7_TRACE* tr2) noexcept nogil

    cpdef float expected_accuracy(self)
    cpdef float score(self, DigitalSequence sequence, Profile profile) except *


cdef class Traces:
    cdef          size_t     _length   # the number of traces in the block
    cdef          size_t     _capacity # the total number of traces that can be stored
    cdef          P7_TRACE** _refs     # the array to pass the sequence references
    cdef          list       _storage  # the actual Python list where `Sequence` objects are stored

    cdef void _allocate(self, size_t n) except *
    cdef void _on_modification(self) noexcept

    cpdef void append(self, Trace trace) except *
    cpdef void clear(self) except *
    cpdef void extend(self, object iterable) except *
    cpdef Trace pop(self, ssize_t index=*)
    cpdef void insert(self, ssize_t index, Trace trace) except *
    cpdef size_t index(self, Trace trace, ssize_t start=*, ssize_t stop=*) except? -1
    cpdef void remove(self, Trace trace) except *


cdef class TraceAligner:
    cdef ESL_SQ** _seqs
    cdef size_t   _nseq

    cpdef Traces compute_traces(self, HMM hmm, DigitalSequenceBlock sequences)
    cpdef MSA align_traces(
        self,
        HMM hmm,
        DigitalSequenceBlock sequences,
        Traces traces,
        bint digitize=*,
        bint trim=*,
        bint all_consensus_cols=*
    )



from . import (
    test_alphabet,
    test_bitfield,
    test_block,
    test_gencode,
    test_keyhash,
    test_matrix,
    test_msa,
    test_msafile,
    test_randomness,
    test_sequence,
    test_sequencefile,
    test_ssi,
    test_vector,
)

def load_tests(loader, suite, pattern):
    suite.addTests(loader.loadTestsFromModule(test_alphabet))
    suite.addTests(loader.loadTestsFromModule(test_bitfield))
    suite.addTests(loader.loadTestsFromModule(test_block))
    suite.addTests(loader.loadTestsFromModule(test_gencode))
    suite.addTests(loader.loadTestsFromModule(test_keyhash))
    suite.addTests(loader.loadTestsFromModule(test_matrix))
    suite.addTests(loader.loadTestsFromModule(test_msa))
    suite.addTests(loader.loadTestsFromModule(test_msafile))
    suite.addTests(loader.loadTestsFromModule(test_randomness))
    suite.addTests(loader.loadTestsFromModule(test_sequence))
    suite.addTests(loader.loadTestsFromModule(test_sequencefile))
    suite.addTests(loader.loadTestsFromModule(test_ssi))
    suite.addTests(loader.loadTestsFromModule(test_vector))
    return suite



# coding: utf-8
# cython: language_level=3, linetrace=True

# --- C imports --------------------------------------------------------------

from libc.stdint cimport uint16_t, uint64_t
from libhmmer.p7_pipeline cimport p7_pipemodes_e

cimport pyhmmer.plan7
from pyhmmer.plan7 cimport TopHits, Pipeline, HMM


# --- Cython classes ---------------------------------------------------------

cdef class Client:

    cdef readonly str            address
    cdef readonly uint16_t       port
    cdef readonly object         socket
    cdef          p7_pipemodes_e mode

    cdef bytearray _recvall(self, size_t message_size)
    cdef TopHits _client(
        self,
        object query,
        uint64_t db,
        list ranges,
        Pipeline pli,
        p7_pipemodes_e mode,
    )


cdef class IterativeSearch(pyhmmer.plan7.IterativeSearch):

    cdef readonly Client   client
    cdef readonly uint64_t db
    cdef          list     ranges
    cdef          dict     options

    cpdef TopHits _search_hmm(self, HMM hmm)



import pickle
import unittest
import sys

from pyhmmer import easel


class TestAlphabet(unittest.TestCase):

    @unittest.skipIf(sys.implementation.name == "pypy", "`getsizeof` not supported on PyPY")
    def test_sizeof(self):
        alphabet = easel.Alphabet.dna()
        self.assertGreater(sys.getsizeof(alphabet), 0)

    def test_is_nucleotide(self):
        self.assertTrue(easel.Alphabet.dna().is_nucleotide())
        self.assertTrue(easel.Alphabet.rna().is_nucleotide())
        self.assertFalse(easel.Alphabet.amino().is_nucleotide())

    def test_is_dna(self):
        self.assertTrue(easel.Alphabet.dna().is_dna())
        self.assertFalse(easel.Alphabet.rna().is_dna())
        self.assertFalse(easel.Alphabet.amino().is_dna())

    def test_is_rna(self):
        self.assertTrue(easel.Alphabet.rna().is_rna())
        self.assertFalse(easel.Alphabet.dna().is_rna())
        self.assertFalse(easel.Alphabet.amino().is_rna())

    def test_is_amino(self):
        self.assertTrue(easel.Alphabet.amino().is_amino())
        self.assertFalse(easel.Alphabet.rna().is_amino())
        self.assertFalse(easel.Alphabet.dna().is_amino())

    def test_eq(self):
        self.assertEqual(easel.Alphabet.dna(), easel.Alphabet.dna())
        self.assertEqual(easel.Alphabet.rna(), easel.Alphabet.rna())
        self.assertEqual(easel.Alphabet.amino(), easel.Alphabet.amino())
        self.assertNotEqual(easel.Alphabet.amino(), easel.Alphabet.dna())
        self.assertNotEqual(easel.Alphabet.amino(), object())

    def test_repr(self):
        self.assertEqual(repr(easel.Alphabet.amino()), "Alphabet.amino()")
        self.assertEqual(repr(easel.Alphabet.rna()), "Alphabet.rna()")
        self.assertEqual(repr(easel.Alphabet.dna()), "Alphabet.dna()")

    def test_K(self):
        self.assertEqual(easel.Alphabet.dna().K, 4)
        self.assertEqual(easel.Alphabet.rna().K, 4)
        self.assertEqual(easel.Alphabet.amino().K, 20)

    def test_picklable(self):
        abc = easel.Alphabet.dna()
        p = pickle.dumps(abc)
        abc2 = pickle.loads(p)
        self.assertIsNot(abc2, abc)
        self.assertEqual(abc, abc)

    def test_encode_decode(self):
        abc = easel.Alphabet.dna()
        seq = "ATGCCCTGACTACA"
        encoded = abc.encode(seq)
        self.assertEqual(len(seq), len(encoded))
        decoded = abc.decode(encoded)
        self.assertEqual(decoded, seq)

    def test_decode_error(self):
        abc = easel.Alphabet.dna()
        seq = bytearray([100, 200])
        self.assertRaises(ValueError, abc.decode, seq)

    def test_encode_error(self):
        abc = easel.Alphabet.dna()
        seq = "123"
        self.assertRaises(ValueError, abc.encode, seq)



import copy
import pickle
import unittest
import sys

from pyhmmer import easel


class TestKeyHash(unittest.TestCase):

    @unittest.skipIf(sys.implementation.name == "pypy", "`getsizeof` not supported on PyPY")
    def test_getsizeof(self):
        kh = easel.KeyHash()
        self.assertGreater(sys.getsizeof(kh), 0)

    def test_len(self):
        kh = easel.KeyHash()
        self.assertEqual(len(kh), 0)
        kh.add(b"first")
        self.assertEqual(len(kh), 1)
        kh.add(b"second")
        self.assertEqual(len(kh), 2)
        kh.add(b"first")
        self.assertEqual(len(kh), 2)
        kh.clear()
        self.assertEqual(len(kh), 0)

    def test_iter(self):
        kh = easel.KeyHash()
        self.assertRaises(StopIteration, next, iter(kh))
        kh.add(b"first")
        kh.add(b"second")
        self.assertEqual(list(iter(kh)), [b"first", b"second"])

    def test_contains(self):
        kh = easel.KeyHash()
        kh.add(b"first")
        self.assertIn(b"first", kh)
        self.assertNotIn(b"second", kh)
        self.assertNotIn(1, kh)

    def test_clear(self):
        kh = easel.KeyHash()
        kh.add(b"first")
        self.assertIn(b"first", kh)
        self.assertEqual(len(kh), 1)
        kh.clear()
        self.assertNotIn(b"first", kh)
        self.assertEqual(len(kh), 0)

    def test_copy(self):
        kh = easel.KeyHash()
        kh.add(b"first")
        k2 = copy.copy(kh)
        self.assertEqual(list(kh), list(k2))
        kh.clear()
        self.assertNotEqual(list(kh), list(k2))

    def test_pickle(self):
        kh = easel.KeyHash()
        kh.add(b"first")
        kh.add(b"second")

        kh2 = pickle.loads(pickle.dumps(kh))
        self.assertEqual(len(kh2), 2)
        self.assertEqual(list(iter(kh2)), [b"first", b"second"])
        self.assertEqual(kh.__getstate__(), kh2.__getstate__())



from libc.stdio cimport FILE
from posix.types cimport off_t

from libeasel.alphabet cimport ESL_ALPHABET
from libhmmer.p7_hmmfile cimport P7_HMMFILE
from libhmmer.impl_neon.p7_oprofile cimport P7_OPROFILE
from libhmmer.impl_neon.p7_omx cimport P7_OM_BLOCK


cdef extern from "impl_neon/impl_neon.h" nogil:

  int p7_oprofile_Write(FILE *ffp, FILE *pfp, P7_OPROFILE *om) except *
  int p7_oprofile_ReadMSV (P7_HMMFILE *hfp, ESL_ALPHABET **byp_abc, P7_OPROFILE **ret_om)
  int p7_oprofile_ReadInfoMSV(P7_HMMFILE *hfp, ESL_ALPHABET **byp_abc, P7_OPROFILE **ret_om)
  int p7_oprofile_ReadBlockMSV(P7_HMMFILE *hfp, ESL_ALPHABET **byp_abc, P7_OM_BLOCK *hmmBlock)
  int p7_oprofile_ReadRest(P7_HMMFILE *hfp, P7_OPROFILE *om)
  int p7_oprofile_Position(P7_HMMFILE *hfp, off_t offset)



from libc.stdint cimport uint8_t, uint16_t, uint32_t


cdef extern from "Python.h":

    ctypedef uint8_t  Py_UCS1
    ctypedef uint16_t Py_UCS2
    ctypedef uint32_t Py_UCS4

    object  PyUnicode_New(Py_ssize_t, Py_UCS4)
    int     PyUnicode_WriteChar(object, Py_ssize_t, Py_UCS4) except -1
    Py_UCS4 PyUnicode_ReadChar(object, Py_ssize_t)

    cdef int PyUnicode_1BYTE_KIND
    cdef int PyUnicode_2BYTE_KIND
    cdef int PyUnicode_4BYTE_KIND

    ssize_t PyUnicode_GET_LENGTH(object)
    int     PyUnicode_KIND(object)
    void*   PyUnicode_DATA(object)
    void    PyUnicode_WRITE(int kind, void *data, Py_ssize_t index, Py_UCS4 value) nogil
    Py_UCS4 PyUnicode_READ (int kind, void *data, Py_ssize_t index) nogil

    Py_UCS1* PyUnicode_1BYTE_DATA(object) nogil
    Py_UCS2* PyUnicode_2BYTE_DATA(object) nogil
    Py_UCS4* PyUnicode_4BYTE_DATA(object) nogil

    object PyUnicode_FromStringAndSize(const char*, Py_ssize_t)
    object PyUnicode_FromKindAndData(int kind, const void *buffer, Py_ssize_t size)

    int    PyUnicode_READY(object) except -1



# coding: utf-8
"""Reimplementation of HMMER binaries with the PyHMMER API.

Note:
    Functions of this module handle parallelization using threads to run
    searches in parallel for the different queries. If less queries are
    given, the number of threads will be reduced to avoid spawning idle
    threads.

"""

from ._hmmalign import hmmalign
from ._hmmpress import hmmpress
from ._hmmscan import hmmscan
from ._hmmsearch import hmmsearch
from ._jackhmmer import jackhmmer
from ._nhmmer import nhmmer
from ._phmmer import phmmer

__all__ = [
    "hmmalign",
    "hmmpress",
    "hmmscan",
    "hmmsearch",
    "jackhmmer",
    "nhmmer",
    "phmmer",
]


# --- Patch for PyPy 3.9 -----------------------------------------------------

cdef extern from *:
    """
    #ifndef HAVE_PYINTERPRETERSTATE_GETID
    int64_t PyInterpreterState_GetID(PyInterpreterState *interp) {
        return 0;
    }
    #endif
    """


from libc.stdint cimport uint64_t
from libc.stdio cimport FILE

from libeasel cimport ESL_DSQ
from libhmmer.p7_profile cimport P7_PROFILE
from libhmmer.p7_gmx cimport P7_GMX


cdef extern from "hmmer.h" nogil:

  int p7_GMSV(const ESL_DSQ *dsq, int L, const P7_PROFILE *gm, P7_GMX *gx, float nu, float *opt_sc)



import copy
import gc
import io
import os
import unittest
import tempfile
from itertools import zip_longest

from pyhmmer import easel

from .. import __name__ as __package__
from .utils import EASEL_FOLDER, resource_files


class TestSequenceFile(unittest.TestCase):

    def test_guess_alphabet_empty_sequence(self):
        buffer = io.BytesIO(b">seq1\n\n")
        self.assertRaises(ValueError, easel.SequenceFile, buffer, format="fasta", digital=True)

    def test_init_error_unknownformat(self):
        with self.assertRaises(ValueError):
            _file = easel.SequenceFile("file.x", format="nonsense")

    def test_init_error_empty(self):
        with tempfile.NamedTemporaryFile() as empty:
            # without a format argument, we can't determine the file format
            self.assertRaises(ValueError, easel.SequenceFile, empty.name)
            # with a format argument, we don't expect the file to be empty
            self.assertRaises(EOFError, easel.SequenceFile, empty.name, "fasta")
        with io.BytesIO(b"") as buffer:
            # without a format argument, we can't determine the file format
            self.assertRaises(ValueError, easel.SequenceFile, buffer)
            # with a format argument, we don't expect the file to be empty
            self.assertRaises(EOFError, easel.SequenceFile, buffer, "fasta")

    def test_init_error_filenotfound(self):
        self.assertRaises(
            FileNotFoundError, easel.SequenceFile, "path/to/missing/file.fa"
        )

    def test_init_error_folder(self):
        folder = tempfile.gettempdir()
        self.assertRaises(IsADirectoryError, easel.SequenceFile, folder)

    @unittest.skipUnless(os.path.exists(EASEL_FOLDER), "test data not available")
    def test_iter(self):
        fasta = os.path.join(EASEL_FOLDER, "formats", "fasta")

        with easel.SequenceFile(fasta) as f:
            seq1, seq2 = list(f)

        self.assertEqual(seq1.name, b"SNRPA_DROME")
        self.assertEqual(seq2.name, b"SNRPA_HUMAN")

    def test_parse(self):
        seq = easel.SequenceFile.parse(b"> EcoRI\nGAATTC\n", format="fasta")
        self.assertEqual(seq.name, b"EcoRI")

    def test_parse_error_unknownformat(self):
        with self.assertRaises(ValueError):
            _file = easel.SequenceFile.parse(b"> EcoRI\nGAATTC\n", format="nonsense")

    @unittest.skipUnless(os.path.exists(EASEL_FOLDER), "test data not available")
    def test_read_skip_info(self):
        fasta = os.path.join(EASEL_FOLDER, "formats", "fasta")

        with easel.SequenceFile(fasta) as f:
            seq = f.read()
            self.assertEqual(seq.name, b"SNRPA_DROME")
            self.assertTrue(seq.sequence.startswith("MEMLPNQTIY"))

        with easel.SequenceFile(fasta) as f:
            seq = f.read(skip_info=True)
            self.assertEqual(seq.name, b"")
            self.assertTrue(seq.sequence.startswith("MEMLPNQTIY"))

    @unittest.skipUnless(os.path.exists(EASEL_FOLDER), "test data not available")
    def test_read_skip_sequence(self):
        fasta = os.path.join(EASEL_FOLDER, "formats", "fasta")

        with easel.SequenceFile(fasta) as f:
            seq = f.read()
            self.assertEqual(seq.name, b"SNRPA_DROME")
            self.assertTrue(seq.sequence.startswith("MEMLPNQTIY"))

        with easel.SequenceFile(fasta) as f:
            seq = f.read(skip_sequence=True)
            self.assertEqual(seq.name, b"SNRPA_DROME")
            self.assertEqual(seq.sequence, "")

    @unittest.skipUnless(resource_files, "importlib.resources.files not available")
    def test_ignore_gaps(self):
        luxc = resource_files(__package__).joinpath("data", "msa", "LuxC.faa")
        # fails if not ignoring gaps (since if contains gaps)
        with easel.SequenceFile(luxc, "fasta") as seq_file:
            self.assertRaises(ValueError, seq_file.read)
        # succeeds if ignoring gaps
        with easel.SequenceFile(luxc, "afa") as seq_file:
            sequences = list(seq_file)
            self.assertEqual(len(sequences), 13)

class _TestReadFilename(object):

    def test_read_filename_guess_format(self):
        # check reading a file without specifying the format works
        for filename, start in zip_longest(self.filenames, self.starts):
            path = os.path.join(self.folder, filename)
            with easel.SequenceFile(path) as f:
                seq = f.read()
                self.assertEqual(seq.sequence[:10], start)

    def test_read_filename_given_format(self):
        # check reading a file while specifying the format works
        for filename, start in zip_longest(self.filenames, self.starts):
            path = os.path.join(self.folder, filename)
            with easel.SequenceFile(path, self.format) as f:
                seq = f.read()
                self.assertEqual(seq.sequence[:10], start)

    def test_read_filename_count_sequences(self):
        # check reading a file while specifying the format works
        for filename, count in zip_longest(self.filenames, self.counts):
            path = os.path.join(self.folder, filename)
            with easel.SequenceFile(path, self.format) as f:
                sequences = list(f)
                self.assertEqual(len(sequences), count)

    def test_read_filename_guess_alphabet(self):
        for filename, alphabet in zip_longest(self.filenames, self.alphabet):
            path = os.path.join(self.folder, filename)
            if alphabet is not None:
                with easel.SequenceFile(path, self.format, digital=True) as f:
                    self.assertEqual(f.alphabet, alphabet)
                # FIXME: Until EddyRivasLab/easel#61 is merged, we cannot
                #        test the case where `guess_alphabet` would throw
                #        an error because of a bug in `sqascii_GuessAlphabet`
                #        causing `eslEOD` to be returned when `eslNOALPHABET`
                #        is expected.

    def test_read_filename_given_alphabet(self):
        for filename, alphabet in zip_longest(self.filenames, self.alphabet):
            path = os.path.join(self.folder, filename)
            if alphabet is not None:
                with easel.SequenceFile(path, self.format, digital=True, alphabet=alphabet) as f:
                    self.assertEqual(f.alphabet, alphabet)

    def test_rewind_filename(self):
        for filename, count, alphabet in zip_longest(self.filenames, self.counts, self.alphabet):
            if alphabet is None:
                continue
            path = os.path.join(self.folder, filename)
            with easel.SequenceFile(path, self.format, digital=True) as f:
                seqs1 = f.read_block()
                f.rewind()
                seqs2 = f.read_block()
                self.assertEqual(len(seqs1), count)
                self.assertEqual(len(seqs2), count)
                self.assertEqual(seqs1, seqs2)


class _TestReadFileObject(object):

    def test_read_fileobject_guess_format(self):
        # check reading a file while specifying the format works
        for filename, start in zip_longest(self.filenames, self.starts):
            path = os.path.join(self.folder, filename)
            with open(path, "rb") as f:
                buffer = io.BytesIO(f.read())
            with easel.SequenceFile(buffer, self.format) as f:
                seq = f.read()
                self.assertEqual(seq.sequence[:10], start)

    def test_read_fileobject_given_format(self):
        # check reading a file while specifying the format works
        for filename, start in zip_longest(self.filenames, self.starts):
            path = os.path.join(self.folder, filename)
            with open(path, "rb") as f:
                buffer = io.BytesIO(f.read())
            with easel.SequenceFile(buffer, self.format) as f:
                seq = f.read()
                self.assertEqual(seq.sequence[:10], start)

    def test_read_fileobject_count_sequences(self):
        # check reading a file while specifying the format works
        for filename, count in zip_longest(self.filenames, self.counts):
            path = os.path.join(self.folder, filename)
            with open(path, "rb") as f:
                buffer = io.BytesIO(f.read())
            with easel.SequenceFile(buffer, self.format) as f:
                sequences = list(f)
                self.assertEqual(len(sequences), count)

    def test_read_fileobject_guess_alphabet(self):
        for filename, alphabet in zip_longest(self.filenames, self.alphabet):
            path = os.path.join(self.folder, filename)
            with open(path, "rb") as f:
                buffer = io.BytesIO(f.read())
            if alphabet is not None:
                with easel.SequenceFile(buffer, self.format, digital=True) as f:
                    self.assertEqual(f.alphabet, alphabet)
                # FIXME: Until EddyRivasLab/easel#61 is merged, we cannot
                #        test the case where `guess_alphabet` would throw
                #        an error because of a bug in `sqascii_GuessAlphabet`
                #        causing `eslEOD` to be returned when `eslNOALPHABET`
                #        is expected.

    def test_rewind_fileobj(self):
        # FIXME: Rewinding a `SequenceFile` with an underlying file-like object
        #        is currently unsupported, because Easel will just reopen the
        #        file
        if self.format in easel.MSAFile._FORMATS:
            raise unittest.SkipTest("{!r} format doesn't support rewinding with file-like objects".format(self.format))
        # check reading a file while specifying the format works
        for filename, count, alphabet in zip_longest(self.filenames, self.counts, self.alphabet):
            if alphabet is None:
                continue
            path = os.path.join(self.folder, filename)
            with open(path, "rb") as f:
                buffer = io.BytesIO(f.read())
            with easel.SequenceFile(buffer, self.format) as f:
                seqs1 = f.read_block()
                f.rewind()
                seqs2 = f.read_block()
                self.assertEqual(len(seqs1), count)
                self.assertEqual(len(seqs2), count)
                self.assertEqual(seqs1, seqs2)


@unittest.skipUnless(os.path.exists(EASEL_FOLDER), "test data not available")
class TestEMBLFile(_TestReadFilename, _TestReadFileObject, unittest.TestCase):
    folder    = os.path.join(EASEL_FOLDER, "formats")
    filenames = ["embl"]
    starts    = ["gaattcctga"]
    format    = "embl"
    counts    = [1]
    alphabet  = [easel.Alphabet.dna()]


@unittest.skipUnless(os.path.exists(EASEL_FOLDER), "test data not available")
class TestFastaFile(_TestReadFilename, _TestReadFileObject, unittest.TestCase):
    folder    = os.path.join(EASEL_FOLDER, "formats")
    filenames = ["fasta", "fasta.2", "fasta.odd.1"]
    starts    = ["MEMLPNQTIY", "MEMLPNQTIY", ""]
    format    = "fasta"
    counts    = [2, 2, 2]
    alphabet  = [easel.Alphabet.amino(), easel.Alphabet.amino(), None]


@unittest.skipUnless(os.path.exists(EASEL_FOLDER), "test data not available")
class TestGenbankFile(_TestReadFilename, _TestReadFileObject, unittest.TestCase):
    folder    = os.path.join(EASEL_FOLDER, "formats")
    filenames = ["genbank", "genbank.2"]
    starts    = ["atccacggcc", "atccacggcc"]
    format    = "genbank"
    counts    = [2, 2]
    alphabet  = [easel.Alphabet.dna(), easel.Alphabet.dna()]


@unittest.skipUnless(os.path.exists(EASEL_FOLDER), "test data not available")
class TestUniprotFile(_TestReadFilename, _TestReadFileObject, unittest.TestCase):
    folder    = os.path.join(EASEL_FOLDER, "formats")
    filenames = ["uniprot"]
    starts    = ["MEMLPNQTIY"]
    format    = "uniprot"
    counts    = [1]
    alphabet  = [easel.Alphabet.amino()]


@unittest.skipUnless(os.path.exists(EASEL_FOLDER), "test data not available")
class TestA2MFile(_TestReadFilename, _TestReadFileObject, unittest.TestCase):
    folder    = os.path.join(EASEL_FOLDER, "esl_msa_testfiles", "a2m")
    filenames = ["a2m.good.1", "a2m.good.2"]
    starts    = ["VLSEGEWQLV", "UCCGAUAUAG"]
    format    = "a2m"
    counts    = [4, 5]
    alphabet  = [easel.Alphabet.amino(), easel.Alphabet.rna()]

    # @unittest.expectedFailure
    def test_read_fileobject_guess_format(self):
        # cannot guess format of A2M sequences, they get identified as AFA
        # or FASTA, and then reading fails because sequences do not have the
        # same length
        super().test_read_fileobject_guess_format()


@unittest.skipUnless(os.path.exists(EASEL_FOLDER), "test data not available")
class TestAlignedFastaFile(_TestReadFilename, _TestReadFileObject, unittest.TestCase):
    folder    = os.path.join(EASEL_FOLDER, "esl_msa_testfiles", "afa")
    filenames = ["afa.good.1", "afa.good.2", "afa.good.3"]
    starts    = ["VLSEGEWQLV", "UCCGAUAUAG", "mqifvktltg"]
    format    = "afa"
    counts    = [4, 5, 6]
    alphabet  = [easel.Alphabet.amino(), easel.Alphabet.rna(), easel.Alphabet.amino()]

    @unittest.expectedFailure
    def test_read_filename_guess_format(self):
        super().test_read_filename_guess_format()


@unittest.skipUnless(os.path.exists(EASEL_FOLDER), "test data not available")
class TestClustalFile(_TestReadFilename, _TestReadFileObject, unittest.TestCase):
    folder    = os.path.join(EASEL_FOLDER, "esl_msa_testfiles", "clustal")
    filenames = ["clustal.good.2"]
    starts    = ["UCCGAUAUAG"]
    format    = "clustal"
    counts    = [5]
    alphabet  = [easel.Alphabet.rna()]


@unittest.skipUnless(os.path.exists(EASEL_FOLDER), "test data not available")
class TestClustalLikeFile(_TestReadFilename, _TestReadFileObject, unittest.TestCase):
    folder    = os.path.join(EASEL_FOLDER, "esl_msa_testfiles", "clustal")
    filenames = ["clustal.good.1"]
    starts    = ["VLSEGEWQLV"]
    format    = "clustallike"
    counts    = [4]
    alphabet  = [easel.Alphabet.amino()]


@unittest.skipUnless(os.path.exists(EASEL_FOLDER), "test data not available")
class TestPhylipFile(_TestReadFilename, _TestReadFileObject, unittest.TestCase):
    folder    = os.path.join(EASEL_FOLDER, "esl_msa_testfiles", "phylip")
    filenames = ["phylip.good.1", "phylip.good.2", "phylip.good.3"]
    starts    = ["AAGCTNGGGC", "ATGGCGAAGG", "MKVILLFVLA"]
    format    = "phylip"
    counts    = [5, 7, 3]
    alphabet  = [easel.Alphabet.dna(), easel.Alphabet.dna(), easel.Alphabet.amino()]


@unittest.skipUnless(os.path.exists(EASEL_FOLDER), "test data not available")
class TestPhylipsFile(_TestReadFilename, _TestReadFileObject, unittest.TestCase):
    folder    = os.path.join(EASEL_FOLDER, "esl_msa_testfiles", "phylips")
    filenames = ["phylips.good.1", "phylips.good.2"]
    starts    = ["AAGCTNGGGC", "MKVILLFVLA"]
    format    = "phylips"
    counts    = [5, 3]
    alphabet  = [easel.Alphabet.dna(), easel.Alphabet.amino()]


@unittest.skipUnless(os.path.exists(EASEL_FOLDER), "test data not available")
class TestPsiblastFile(_TestReadFilename, _TestReadFileObject, unittest.TestCase):
    folder    = os.path.join(EASEL_FOLDER, "esl_msa_testfiles", "psiblast")
    filenames = ["psiblast.good.1", "psiblast.good.2"]
    starts    = ["VLSEGEWQLV", "UCCGAUAUAG"]
    format    = "psiblast"
    counts    = [4, 5]
    alphabet  = [easel.Alphabet.amino(), easel.Alphabet.rna()]


@unittest.skipUnless(os.path.exists(EASEL_FOLDER), "test data not available")
class TestSelexFile(_TestReadFilename, _TestReadFileObject, unittest.TestCase):
    folder    = os.path.join(EASEL_FOLDER, "esl_msa_testfiles", "selex")
    filenames = ["selex.good.1", "selex.good.2", "selex.good.3", "selex.good.4"]
    starts    = ["ACDEFGHIKL", "ACDEFGHIKL", "ACDEFGHIKL", "gGAGUAAGAU"]
    format    = "selex"
    counts    = [5, 5, 7, 11]
    alphabet  = [easel.Alphabet.amino(), easel.Alphabet.amino(), easel.Alphabet.amino(), easel.Alphabet.rna()]

    @unittest.expectedFailure
    def test_read_fileobject_given_format(self):
        # unknown error: '#=RF line isn't in expected order in block'
        super().test_read_fileobject_given_format()

    @unittest.expectedFailure
    def test_read_fileobject_guess_format(self):
        # unknown error: '#=RF line isn't in expected order in block'
        super().test_read_fileobject_guess_format()

    @unittest.expectedFailure
    def test_read_fileobject_count_sequences(self):
        # unknown error: '#=RF line isn't in expected order in block'
        super().test_read_fileobject_count_sequences()


@unittest.skipUnless(os.path.exists(EASEL_FOLDER), "test data not available")
class TestStockholmFile(_TestReadFilename, _TestReadFileObject, unittest.TestCase):
    folder    = os.path.join(EASEL_FOLDER, "esl_msa_testfiles", "stockholm")
    filenames = ["stockholm.good.1"]
    starts    = ["ACDEFGHKLM"]
    format    = "stockholm"
    counts    = [7]
    alphabet  = [easel.Alphabet.amino()]



import abc
import ctypes
import collections
import contextlib
import itertools
import multiprocessing.synchronize
import multiprocessing.connection
import multiprocessing.resource_sharer
import operator
import queue
import sys
import typing
import threading

from ..easel import DigitalSequenceBlock, SequenceFile, DigitalSequence, DigitalMSA
from ..plan7 import Pipeline, Builder, HMMPressedFile, HMM, Profile, OptimizedProfile, IterationResult, OptimizedProfileBlock
from ..utils import singledispatchmethod

# --- Type annotations ---------------------------------------------------------

# the query type for the pipeline
_Q = typing.TypeVar("_Q")
# the target type for the pipeline
_T = typing.TypeVar("_T")
# the result type for the pipeline
_R = typing.TypeVar("_R")

# type aliases
_AnyProfile = typing.Union[HMM, Profile, OptimizedProfile]

# `typing.Literal`` is only available in Python 3.8 and later
if typing.TYPE_CHECKING:

    if sys.version_info >= (3, 8):
        from typing import Literal, TypedDict
    else:
        from typing_extensions import Literal, TypedDict  # type: ignore

    if sys.version_info >= (3, 11):
        from typing import Unpack
    else:
        from typing_extensions import Unpack

    from .plan7 import BIT_CUTOFFS, STRAND, Background
    from .easel import Alphabet

    BACKEND = Literal["threading", "multiprocessing"]
    PARALLEL = Literal["queries", "targets"]

    class PipelineOptions(TypedDict, total=False):
        alphabet: Alphabet
        background: typing.Optional[Background]
        bias_filter: bool
        null2: bool
        seed: int
        Z: float
        domZ: typing.Optional[float]
        F1: float
        F2: float
        F3: float
        E: float
        T: typing.Optional[float]
        domE: float
        domT: typing.Optional[float]
        incE: float
        incT: typing.Optional[float]
        incdomE: float
        incdomT: typing.Optional[float]
        bit_cutoffs: typing.Optional[BIT_CUTOFFS]

    class LongTargetsPipelineOptions(PipelineOptions, total=False):
        strand: typing.Optional[STRAND]
        B1: int
        B2: int
        B3: int
        block_length: int
        window_length: typing.Optional[int]
        window_beta: typing.Optional[float]


# --- Result class -------------------------------------------------------------

class _BaseChore(typing.Generic[_Q, _R], abc.ABC):
    query: _Q

    def __init__(self, query: _Q):
        self.query = query

    # --- Dispatcher methods ---

    @abc.abstractmethod
    def available(self) -> bool:
        """Return whether the chore is done and results are available."""

    @abc.abstractmethod
    def wait(self, timeout: typing.Optional[float] = None) -> bool:
        """Wait for the chore to be done."""

    @abc.abstractmethod
    def get(self) -> _R:
        """Get the results of the chore, blocking if the chore was not done."""

    # --- Worker methods ---

    @abc.abstractmethod
    def complete(self, result: _R) -> None:
        """Mark the chore as done and record ``result`` as the results."""

    @abc.abstractmethod
    def fail(self, exception: BaseException) -> None:
        """Mark the chore as done and record ``exception`` as the error."""


class _ThreadChore(typing.Generic[_Q, _R], _BaseChore[_Q, _R]):
    """A chore for a worker thread.

    Attributes:
        query (`object`): The query object to be processed by the worker
            thread. Exact type depends on the pipeline type.
        event (`threading.Event`): An event flag to set when the query
            is done being processed.
        result (`pyhmmer.plan7.TopHits` or `BaseException`): The results
            obtained after processing the query, or an exception on failure.

    """

    result: typing.Union[_R, BaseException, None]
    event: threading.Event

    def __init__(self, query: _Q) -> None:
        """Create a new chore from the given query."""
        super().__init__(query)
        self.event = threading.Event()
        self.result = None

    def available(self) -> bool:
        return self.event.is_set()

    def wait(self, timeout: typing.Optional[float] = None) -> bool:
        return self.event.wait(timeout)

    def get(self) -> _R:
        self.event.wait()
        assert self.result is not None
        if isinstance(self.result, BaseException):
            raise self.result
        else:
            return self.result

    def complete(self, result: _R) -> None:
        self.result = result
        self.event.set()

    def fail(self, exception: BaseException) -> None:
        self.result = exception
        self.event.set()


class _ProcessChore(typing.Generic[_Q, _R], _BaseChore[_Q, _R]):
    """A chore for a worker process.

    Attributes:
        query (`object`): The query object to be processed by the worker
            thread. Exact type depends on the pipeline type.
        conns (`multiprocessing.Connection`): The pipe end where the
            worker should send the result.
        connr (`multiprocessing.Connection`): The pipe end where the
            dispatcher should receive the result.

    """

    connr: multiprocessing.connection.Connection
    conns: multiprocessing.connection.Connection

    def __init__(self, query: _Q) -> None:
        """Create a new chore from the given query."""
        super().__init__(query)
        self.connr, self.conns = multiprocessing.Pipe()

    def available(self) -> bool:
        """Return whether the chore is done and results are available."""
        return self.connr.poll(0)

    def wait(self, timeout: typing.Optional[float] = None) -> bool:
        """Wait for the chore to be done."""
        return self.connr.poll(timeout)

    def complete(self, result: _R) -> None:
        """Mark the chore as done and record ``result`` as the results."""
        self.conns.send(result)

    def fail(self, exception: BaseException) -> None:
        """Mark the chore as done and record ``exception`` as the error."""
        self.conns.send(exception)

    def get(self) -> _R:
        """Get the results of the chore, blocking if the chore was not done."""
        self.wait()
        result = self.connr.recv()
        self.connr.close()
        self.conns.close()
        assert result is not None
        if isinstance(result, BaseException):
            raise result
        else:
            return result  # type: ignore


# --- Workers ------------------------------------------------------------------

class _BaseWorker(typing.Generic[_Q, _T, _R]):
    """A generic worker to parallelize a pipelined search.

    Attributes:
        targets (`DigitalSequenceBlock` or `OptimizedProfileBlock`): The
            target to search for hits, either a digital sequence block while
            in search mode, or an optimized profile block while in scan mode.
        query_queue (`queue.Queue`): The queue used to pass queries between
            the dispatcher and the workers. It contains the query, its index
            so that the results can be returned in the same order, and
            a way to pass back the results corresponding to that query
            back to the dispatcher.
        query_count (`multiprocessing.Value`): An atomic counter storing
            the total number of queries that have been loaded so far.
            Passed to the ``callback`` so that an UI can show the total
            for a progress bar.
        kill_switch (`threading.Event`): An event flag shared between
            all worker threads, used to notify emergency exit.
        callback (`callable`, optional): An optional callback to be called
            after each query has been processed. It should accept two
            arguments: the query object that was processed, and the total
            number of queries read until now. For `multiprocessing` workers,
            this callback should be picklable.
        options (`dict`): A dictionary of options to be passed to the
            `pyhmmer.plan7.Pipeline` object wrapped by the worker.
        builder (`~pyhmmer.plan7.Builder`, *optional*): The builder to use
            for translating sequence or alignment queries into `HMM` objects.
            May be `None` if the queries are expected to be `HMM` only.

    """

    pipeline_class: typing.ClassVar[typing.Type[Pipeline]] = Pipeline

    @staticmethod
    def _none_callback(hmm: _Q, total: int) -> None:
        pass

    def __init__(
        self,
        targets: _T,
        query_queue: "queue.Queue[typing.Optional[_BaseChore[_Q, _R]]]",
        query_count: multiprocessing.Value,  # type: ignore
        kill_switch: threading.Event,
        callback: typing.Optional[typing.Callable[[_Q, int], None]],
        options: "PipelineOptions",
        builder: typing.Optional[Builder] = None,
    ) -> None:
        super().__init__()
        self.options = options
        self.targets: _T = targets
        self.pipeline: typing.Optional[Pipeline] = None
        self.pipeline_options = options
        self.query_queue: "queue.Queue[typing.Optional[_BaseChore[_Q, _R]]]" = query_queue
        self.query_count = query_count
        self.callback: typing.Optional[typing.Callable[[_Q, int], None]] = (
            callback or self._none_callback
        )
        self.kill_switch = kill_switch
        self.builder = builder

    def run(self) -> None:
        while not self.is_killed():
            # attempt to get the next argument, with a timeout
            # so that the worker can periodically check if it has
            # been killed, even when no queries are available
            try:
                chore = self.query_queue.get(timeout=1)
            except queue.Empty:
                continue
            except BrokenPipeError:
                self.kill()
                break
            except ConnectionError:
                self.kill()
                break
            # check if arguments from the queue are a poison-pill (`None`),
            # in which case the thread will stop running
            if chore is None:
                break
            # process the query, making sure to capture any exception
            # and then mark the hits as "found" using a `threading.Event`
            try:
                hits = self.process(chore.query)
                chore.complete(hits)
            except BaseException as exc:
                self.kill()
                chore.fail(exc)
        if isinstance(self.targets, (SequenceFile, HMMPressedFile)):
            self.targets.close()

    if typing.TYPE_CHECKING:
        def start(self) -> None: ...

    def is_killed(self) -> bool:
        try:
            return self.kill_switch.is_set()
        except BrokenPipeError:  # the connection was closed already
            return True
        except ConnectionError:  # the connection was closed already
            return True
        except FileNotFoundError:  # the Event manager has been closed already
            return True

    def kill(self) -> None:
        """Set the synchronized kill switch for all threads."""
        if not self.is_killed():
            self.kill_switch.set()

    def process(self, query: _Q) -> _R:
        """Process a single query and return the resulting hits."""
        if isinstance(self.targets, (HMMPressedFile, SequenceFile)):
            self.targets.rewind()
        if self.pipeline is None:
            self.pipeline = self.pipeline_class(**self.pipeline_options)
        hits = self.query(query)
        self.callback(query, self.query_count.value)  # type: ignore
        self.pipeline.clear()
        return hits

    @abc.abstractmethod
    def query(self, query: _Q) -> _R:
        """Run a single query against the target database."""
        return NotImplemented


# --- Dispatcher ---------------------------------------------------------------

class _BaseDispatcher(typing.Generic[_Q, _T, _R], abc.ABC):
    def __init__(
        self,
        queries: typing.Iterable[_Q],
        targets: _T,
        cpus: int = 0,
        callback: typing.Optional[typing.Callable[[_Q, int], None]] = None,
        builder: typing.Optional[Builder] = None,
        timeout: int = 1,
        backend: "BACKEND" = "threading",
        **options,  # type: Unpack[PipelineOptions]
    ) -> None:
        self.queries = queries
        self.targets: _T = targets
        self.callback: typing.Optional[typing.Callable[[_Q, int], None]] = callback
        self.options = options
        self.builder = builder
        self.timeout = timeout

        # make sure a positive number of CPUs is requested
        if cpus <= 0:
            raise ValueError("`cpus` must be strictly positive, not {!r}".format(cpus))

        # reduce the number of threads if there are less queries (at best
        # use one thread by query)
        hint = operator.length_hint(queries, -1)
        self.cpus = 1 if hint == 0 else min(cpus, hint) if hint > 0 else cpus

        # force "threading" backend when running everything in main thread
        self.backend = "threading" if self.cpus == 1 else backend

    @abc.abstractmethod
    def _new_worker(
        self,
        query_queue: "queue.Queue[typing.Optional[_BaseChore[_Q, _R]]]",
        query_count: "multiprocessing.Value[int]",  # type: ignore
        kill_switch: threading.Event,
    ) -> _BaseWorker[_Q, _T, _R]:
        return NotImplemented

    def _new_chore(
        self,
        query: _Q,
    ) -> _BaseChore[_Q, _R]:
        if self.backend == "threading":
            return _ThreadChore(query)
        elif self.backend == "multiprocessing":
            return _ProcessChore(query)
        else:
            raise ValueError(f"Invalid parallel backend: {self.backend!r}")

    def _single_threaded(self) -> typing.Iterator[_R]:
        # create the queues to pass the HMM objects around, as well as atomic
        # values that we use to synchronize the threads
        query_queue = queue.Queue()  # type: ignore
        query_count = multiprocessing.Value(ctypes.c_ulong)
        kill_switch = threading.Event()

        # create the thread (to recycle code)
        thread = self._new_worker(query_queue, query_count, kill_switch)

        # process each HMM iteratively and yield the result
        # immediately so that the user can iterate over the
        # TopHits one at a time
        for query in self.queries:
            query_count.value += 1
            yield thread.process(query)

        # close the targets if they were coming from a file
        if isinstance(thread.targets, (SequenceFile, HMMPressedFile)):
            thread.targets.close()

    def _multi_threaded(self) -> typing.Iterator[_R]:
        with contextlib.ExitStack() as ctx:
            # create the queues to pass the query objects around, as well as
            # atomic values that we use to synchronize the threads
            results: typing.Deque[_BaseChore[_Q, _R]] = collections.deque()
            if self.backend == "multiprocessing":
                manager = ctx.enter_context(multiprocessing.Manager())
                query_queue = ctx.enter_context(contextlib.closing(multiprocessing.Queue(maxsize=2*self.cpus)))
                query_count = manager.Value(ctypes.c_ulong, 0)
                kill_switch = manager.Event()
            elif self.backend == "threading":
                query_queue = queue.Queue(maxsize=2*self.cpus)
                query_count = multiprocessing.Value(ctypes.c_ulong)  # type: ignore
                kill_switch = threading.Event()

            # create and launch one pipeline thread per CPU
            workers = []
            for _ in range(self.cpus):
                worker = self._new_worker(query_queue, query_count, kill_switch)
                worker.start()
                workers.append(worker)

            # catch exceptions to kill threads in the background before exiting
            try:
                # alternate between feeding queries to the threads and
                # yielding back results, if available. the priority is
                # given to filling the query queue, so that no worker
                # ever idles.
                for query in self.queries:
                    # prepare to add the next query to the queue
                    query_count.value += 1
                    chore = self._new_chore(query)
                    # attempt to add the query to the query queue,
                    # but use a timeout so that it doesn't deadlock
                    # if the background thread errored (which would
                    # set the kill switch), typically because the
                    # user-provided callback failed
                    while not kill_switch.is_set():
                        with contextlib.suppress(queue.Full):
                            query_queue.put(chore, timeout=self.timeout)  # <-- blocks if too many chores in queue
                            results.append(chore)
                            break
                    # aggressively wait for the result with a very short
                    # timeout, and exit the loop if the queue is not full
                    if results[0].available():
                        yield results[0].get()
                        results.popleft()
                # now that we exhausted all queries, poison pill the
                # threads so they stop on their own gracefully
                for _ in workers:
                    query_queue.put(None)
                # wait for final workers
                for worker in workers:
                    worker.join()
                if self.backend == "multiprocessing":
                    worker.query_queue.close()
                    worker.query_queue.join_thread()
                # yield all remaining results, in order
                while results:
                    result = results[0].get()  # <-- blocks until result is available
                    results.popleft()
                    yield result
            except BaseException as e:
                # make sure threads are killed to avoid being stuck,
                # e.g. after a KeyboardInterrupt, then re-raise
                try:
                    kill_switch.set()
                except queue.Full:
                    pass
                for worker in workers:
                    worker.join()
                    if self.backend == "multiprocessing":
                        worker.query_queue.close()
                raise e

    def run(self) -> typing.Iterator[_R]:
        if self.cpus == 1:
            return self._single_threaded()
        else:
            return self._multi_threaded()



from . import (
    test_alignment,
    test_background,
    test_block,
    test_builder,
    test_hit,
    test_hmm,
    test_hmmfile,
    test_optimizedprofile,
    test_pipeline,
    test_profile,
    test_tophits,
    test_tracealigner,
    test_traces,
)

def load_tests(loader, suite, pattern):
    suite.addTests(loader.loadTestsFromModule(test_alignment))
    suite.addTests(loader.loadTestsFromModule(test_background))
    suite.addTests(loader.loadTestsFromModule(test_block))
    suite.addTests(loader.loadTestsFromModule(test_builder))
    suite.addTests(loader.loadTestsFromModule(test_hit))
    suite.addTests(loader.loadTestsFromModule(test_hmm))
    suite.addTests(loader.loadTestsFromModule(test_hmmfile))
    suite.addTests(loader.loadTestsFromModule(test_optimizedprofile))
    suite.addTests(loader.loadTestsFromModule(test_pipeline))
    suite.addTests(loader.loadTestsFromModule(test_profile))
    suite.addTests(loader.loadTestsFromModule(test_tophits))
    suite.addTests(loader.loadTestsFromModule(test_tracealigner))
    suite.addTests(loader.loadTestsFromModule(test_traces))
    return suite



import io
import itertools
import os
import pickle
import shutil
import unittest
import tempfile

import pyhmmer
from pyhmmer.errors import EaselError
from pyhmmer.easel import SequenceFile, TextMSA, DigitalMSA
from pyhmmer.plan7 import HMMFile, Pipeline, TopHits

from .. import __name__ as __package__
from .utils import HMMER_FOLDER, resource_files


@unittest.skipUnless(resource_files, "importlib.resources.files not available")
class TestTopHits(unittest.TestCase):

    def assertAlignmentEqual(self, a1, a2):
        for attr in (
            "hmm_from",
            "hmm_to",
            "hmm_name",
            "hmm_accession",
            "hmm_sequence",
            "posterior_probabilities",
            "target_from",
            "target_to",
            "target_name",
            "target_sequence",
            "identity_sequence",
        ):
            self.assertEqual(
                getattr(a1, attr),
                getattr(a2, attr),
                "attribute {!r} differs".format(attr)
            )

    def assertDomainEqual(self, d1, d2):
        for attr in (
            "env_from",
            "env_to",
            "score",
            "bias",
            "correction",
            "envelope_score",
            "c_evalue",
            "i_evalue",
            "pvalue",
        ):
            self.assertEqual(
                getattr(d1, attr),
                getattr(d2, attr),
                "attribute {!r} differs".format(attr)
            )
        self.assertAlignmentEqual(d1.alignment, d2.alignment)

    def assertHitEqual(self, h1, h2):
        for attr in (
            "name",
            "accession",
            "description",
            "score",
            "pre_score",
            "sum_score",
            "bias",
            "evalue",
            "pvalue",
        ):
            self.assertEqual(
                getattr(h1, attr),
                getattr(h2, attr),
                "attribute {!r} differs".format(attr)
            )
        for d1, d2 in zip(h1.domains, h2.domains):
            self.assertDomainEqual(d1, d2)
        self.assertEqual(len(h1.domains), len(h2.domains))
        self.assertEqual(len(h1.domains.reported), len(h2.domains.reported))
        self.assertEqual(len(h1.domains.included), len(h2.domains.included))
        self.assertEqual(sum(d.reported for d in h1.domains), sum(d.reported for d in h2.domains))
        self.assertEqual(sum(d.included for d in h1.domains), sum(d.included for d in h2.domains))

    def assertHitsEqual(self, hits1, hits2):
        self.assertEqual(hits1.query.name, hits2.query.name)
        self.assertEqual(hits1.query.accession, hits2.query.accession)
        self.assertEqual(hits1.query, hits2.query)
        self.assertEqual(len(hits1), len(hits2))
        self.assertEqual(len(hits1.included), len(hits2.included))
        self.assertEqual(len(hits1.reported), len(hits2.reported))
        for h1, h2 in zip(hits1, hits2):
            self.assertHitEqual(h1, h2)

    @classmethod
    def setUpClass(cls):
        hmm_path = resource_files(__package__).joinpath("data", "hmms", "txt", "PF02826.hmm")
        seqs_path = resource_files(__package__).joinpath("data", "seqs", "938293.PRJEB85.HG003687.faa")

        with HMMFile(hmm_path) as hmm_file:
            cls.hmm = hmm_file.read()
        with SequenceFile(seqs_path, digital=True, alphabet=cls.hmm.alphabet) as seqs_file:
            cls.seqs = seqs_file.read_block()

        cls.pipeline = Pipeline(alphabet=cls.hmm.alphabet)
        cls._hits = cls.pipeline.search_hmm(cls.hmm, cls.seqs)

    def setUp(self):
        self.hits = self._hits.copy()

    def test_mode(self):
        pipeline = Pipeline(alphabet=self.hmm.alphabet)
        search_hits = pipeline.search_hmm(self.hmm, self.seqs)
        self.assertEqual(search_hits.mode, "search")

    def test_bool(self):
        self.assertFalse(pyhmmer.plan7.TopHits(self.hmm))
        self.assertTrue(self.hits)

    def test_index_error(self):
        # check out of bounds indexing is caught
        with self.assertRaises(IndexError):
            self.hits[len(self.hits) + 1]
        # check out of bounds negative indexing works
        with self.assertRaises(IndexError):
            self.hits[-len(self.hits) - 1]

    def test_index_negative(self):
        dom = self.hits[len(self.hits) - 1]
        dom_last = self.hits[-1]
        self.assertEqual(dom.name, dom_last.name)

    def test_Z(self):
        empty = TopHits(self.hmm)
        self.assertEqual(empty.Z, 0)
        self.assertEqual(self.hits.Z, len(self.seqs))

    def test_strand(self):
        empty = TopHits(self.hmm)
        self.assertIs(empty.strand, None)

    def test_searched_sequences(self):
        empty = TopHits(self.hmm)
        self.assertEqual(empty.searched_sequences, 0)
        self.assertEqual(self.hits.searched_sequences, len(self.seqs))

    def test_searched_nodes(self):
        empty = TopHits(self.hmm)
        self.assertEqual(empty.searched_nodes, 0)
        self.assertEqual(self.hits.searched_nodes, self.hmm.M)

    def test_searched_residues(self):
        empty = TopHits(self.hmm)
        self.assertEqual(empty.searched_residues, 0)
        self.assertEqual(self.hits.searched_residues, sum(map(len, self.seqs)))

    def test_searched_models(self):
        empty = TopHits(self.hmm)
        self.assertEqual(empty.searched_sequences, 0)
        self.assertEqual(self.hits.searched_models, 1)

    def test_merge_empty(self):
        empty = TopHits(self.hmm)
        self.assertFalse(empty.long_targets)
        self.assertEqual(empty.Z, 0.0)
        self.assertEqual(empty.domZ, 0.0)

        empty2 = empty.merge()
        self.assertFalse(empty2.long_targets)
        self.assertEqual(empty2.Z, 0.0)
        self.assertEqual(empty2.domZ, 0.0)

        merged_empty = empty.merge(TopHits(self.hmm))
        self.assertHitsEqual(merged_empty, empty)
        self.assertEqual(merged_empty.searched_residues, 0)
        self.assertEqual(merged_empty.searched_sequences, 0)
        self.assertFalse(merged_empty.long_targets)
        self.assertFalse(self.hits.long_targets)
        self.assertEqual(merged_empty.Z, 0.0)
        self.assertEqual(merged_empty.domZ, 0.0)

        merged = empty.merge(self.hits)
        self.assertEqual(merged.query.name, self.hits.query.name)
        self.assertEqual(merged.query.M, self.hits.query.M)
        self.assertEqual(merged.query.accession, self.hits.query.accession)
        self.assertEqual(merged.Z, self.hits.Z)
        self.assertEqual(merged.E, self.hits.E)
        self.assertHitsEqual(merged, self.hits)

    def test_merge_pipeline(self):
        pipeline = Pipeline(alphabet=self.hmm.alphabet)

        hits1 = pipeline.search_hmm(self.hmm, self.seqs[:1000])
        self.assertEqual(hits1.Z, 1000)
        hits2 = pipeline.search_hmm(self.hmm, self.seqs[1000:2000])
        self.assertEqual(hits2.Z, 1000)
        hits3 = pipeline.search_hmm(self.hmm, self.seqs[2000:])
        self.assertEqual(hits3.Z, len(self.seqs) - 2000)

        merged = hits1.merge(hits2, hits3)
        self.assertEqual(merged.mode, "search")

        hits = pipeline.search_hmm(self.hmm, self.seqs)
        self.assertEqual(len(hits1) + len(hits2) + len(hits3), len(hits))
        self.assertEqual(len(merged), len(hits))

        self.assertEqual(merged.searched_sequences, hits.searched_sequences)
        self.assertEqual(merged.Z, hits.Z)
        self.assertEqual(merged.domZ, hits.domZ)
        self.assertIs(merged.query, hits.query)
        self.assertEqual(merged.query.name, hits.query.name)
        self.assertEqual(merged.query.M, hits.query.M)
        self.assertEqual(merged.query.accession, hits.query.accession)
        self.assertEqual(merged.E, hits.E)
        self.assertEqual(merged.domE, hits.domE)

        for hit in merged:
            nincluded = sum(domain.included for domain in hit.domains)
            self.assertEqual(nincluded, len(hit.domains.included))
            nreported = sum(domain.reported for domain in hit.domains)
            self.assertEqual(nreported, len(hit.domains.reported))

        self.assertHitsEqual(merged, hits)

        # number of reported & included hits in the final merged hits may be
        # different of the sum here, because we used E-value thresholds, which
        # are affected by how many hits were effectively reported

    def test_merge_pipeline_byscore(self):
        pipeline = Pipeline(alphabet=self.hmm.alphabet, T=10.0, domT=10.0, incT=10.0, incdomT=10.0)

        hits1 = pipeline.search_hmm(self.hmm, self.seqs[:1000])
        self.assertEqual(hits1.Z, 1000)
        hits2 = pipeline.search_hmm(self.hmm, self.seqs[1000:2000])
        self.assertEqual(hits2.Z, 1000)
        hits3 = pipeline.search_hmm(self.hmm, self.seqs[2000:])
        self.assertEqual(hits3.Z, len(self.seqs) - 2000)

        merged = hits1.merge(hits2, hits3)
        self.assertEqual(merged.mode, "search")

        hits = pipeline.search_hmm(self.hmm, self.seqs)
        self.assertEqual(len(hits1) + len(hits2) + len(hits3), len(hits))
        self.assertEqual(len(merged), len(hits))

        self.assertEqual(merged.searched_sequences, hits.searched_sequences)
        self.assertEqual(merged.Z, hits.Z)
        self.assertEqual(merged.domZ, hits.domZ)
        self.assertIs(merged.query, hits.query)
        self.assertEqual(merged.query.name, hits.query.name)
        self.assertEqual(merged.query.M, hits.query.M)
        self.assertEqual(merged.query.accession, hits.query.accession)
        self.assertEqual(merged.E, hits.E)
        self.assertEqual(merged.domE, hits.domE)

        for hit in merged:
            nincluded = sum(domain.included for domain in hit.domains)
            self.assertEqual(nincluded, len(hit.domains.included))
            nreported = sum(domain.reported for domain in hit.domains)
            self.assertEqual(nreported, len(hit.domains.reported))

        self.assertHitsEqual(merged, hits)

        # number of reported & included hits in the final merged hits should
        # be equal to the sum of reported & included hits in each sub hits,
        # because we used bitscore thresholds, which are independent of
        # the number of hits

        reported = lambda hits: sum(len(hit.domains.reported) for hit in hits)
        included = lambda hits: sum(len(hit.domains.reported) for hit in hits)
        hits_nreported = reported(hits1) + reported(hits2) + reported(hits3)
        hits_nincluded = included(hits1) + included(hits2) + included(hits3)
        self.assertEqual(reported(merged), hits_nreported)
        self.assertEqual(included(merged), hits_nincluded)

    def test_merged_pipeline_fixed_Z(self):
        pipeline = Pipeline(alphabet=self.hmm.alphabet, Z=200.0)
        hits1 = pipeline.search_hmm(self.hmm, self.seqs[:1000])
        hits2 = pipeline.search_hmm(self.hmm, self.seqs[1000:2000])
        hits3 = pipeline.search_hmm(self.hmm, self.seqs[2000:])

        self.assertEqual(hits1.Z, 200.0)
        self.assertEqual(hits2.Z, 200.0)
        self.assertEqual(hits3.Z, 200.0)

        merged = hits1.merge(hits2, hits3)
        self.assertEqual(merged.Z, 200.0)

        hits = pipeline.search_hmm(self.hmm, self.seqs)
        self.assertHitsEqual(merged, hits)

    def test_copy(self):
        copy = self.hits.copy()
        self.assertHitsEqual(copy, self.hits)

    def test_sort(self):
        # check the hits are sorted by default
        self.assertTrue(self.hits.is_sorted())

        # check sorting
        self.hits.sort()
        self.assertTrue(self.hits.is_sorted())
        self.assertFalse(self.hits.is_sorted(by="seqidx"))

        # check sorting using a different key
        self.hits.sort(by="seqidx")
        self.assertTrue(self.hits.is_sorted(by="seqidx"))
        self.assertFalse(self.hits.is_sorted(by="key"))

        # check error otherwise
        self.assertRaises(ValueError, self.hits.sort, by="nonsense")
        self.assertRaises(ValueError, self.hits.is_sorted, by="nonsense")

    def test_sort_while_hit(self):
        # check sorting while a Hit instance exists does not crash, and that
        # the Hit still points to the right data
        hit = self.hits[-1]
        self.assertEqual(hit.name, b"938293.PRJEB85.HG003687_187")
        self.hits.sort(by="seqidx")
        if self.hits[-1].name != b"938293.PRJEB85.HG003687_187":
            self.assertEqual(hit.name, b"938293.PRJEB85.HG003687_187")

    def test_threshold(self):
        # after thresholding, one of the hits will not be included.
        self.assertEqual(len(self.hits.included), 15)
        self.assertEqual(len(self.hits.reported), 22)

    def test_to_msa(self):
        msa = self.hits.to_msa(self.hmm.alphabet)
        self.assertIsInstance(msa, TextMSA)
        unique_names = { seq.name.split(b"/")[0] for seq in msa.sequences }
        self.assertEqual(len(unique_names), len(self.hits.included))

        msa_d = self.hits.to_msa(
            self.hmm.alphabet,
            trim=True,
            digitize=True,
            all_consensus_cols=True
        )
        self.assertIsInstance(msa_d, DigitalMSA)

    def test_pickle(self):
        pickled = pickle.loads(pickle.dumps(self.hits))
        self.assertHitsEqual(pickled, self.hits)

    def test_query(self):
        self.assertIs(self.hits.query, self.hmm)

    def test_query_name(self):
        self.assertEqual(self.hits.query.name, self.hmm.name)

    def test_query_accession(self):
        self.assertEqual(self.hits.query.accession, self.hmm.accession)

    def test_query_length(self):
        self.assertEqual(self.hits.query.M, self.hmm.M)

    def test_write_target(self):
        buffer = io.BytesIO()
        self.hits.write(buffer, format="targets")
        lines = buffer.getvalue().decode().splitlines()

        with resource_files(__package__).joinpath("data", "tables", "PF02826.tbl").open() as f:
            expected = f.read().splitlines()
        while expected[-1].startswith("#"):
            expected.pop()

        self.assertMultiLineEqual("\n".join(lines), "\n".join(expected))

    def test_write_domains(self):
        buffer = io.BytesIO()
        self.hits.write(buffer, format="domains")
        lines = buffer.getvalue().decode().splitlines()

        with resource_files(__package__).joinpath("data", "tables", "PF02826.domtbl").open() as f:
            expected = f.read().splitlines()
        while expected[-1].startswith("#"):
            expected.pop()

        self.assertMultiLineEqual("\n".join(lines), "\n".join(expected))

    def test_manual_report(self):
        self.assertEqual(len(self.hits.reported), 22)
        self.assertTrue(self.hits[0].reported)

        self.hits[0].reported = False
        self.assertEqual(len(self.hits.reported), 21)
        self.assertFalse(self.hits[0].reported)

        self.hits[0].reported = True
        self.assertEqual(len(self.hits.reported), 22)
        self.assertTrue(self.hits[0].reported)

    def test_manual_inclusion(self):
        self.assertEqual(len(self.hits.included), 15)
        self.assertEqual(len(self.hits.reported), 22)
        self.assertTrue(self.hits[0].included)
        self.assertTrue(self.hits[0].reported)

        self.hits[0].included = False
        self.assertEqual(len(self.hits.included), 14)
        self.assertEqual(len(self.hits.reported), 22)
        self.assertFalse(self.hits[0].included)
        self.assertTrue(self.hits[0].reported)

        self.hits[0].included = True
        self.assertEqual(len(self.hits.included), 15)
        self.assertEqual(len(self.hits.reported), 22)
        self.assertTrue(self.hits[0].included)
        self.assertTrue(self.hits[0].reported)

    def test_manual_drop(self):
        self.assertEqual(len(self.hits.included), 15)
        self.assertEqual(len(self.hits.reported), 22)
        self.assertTrue(self.hits[0].included)
        self.assertTrue(self.hits[0].reported)

        self.hits[0].dropped = True
        self.assertEqual(len(self.hits.included), 14)
        self.assertEqual(len(self.hits.reported), 22)
        self.assertTrue(self.hits[0].dropped)
        self.assertFalse(self.hits[0].included)
        self.assertTrue(self.hits[0].reported)

        self.hits[0].dropped = False
        self.assertEqual(len(self.hits.included), 14)
        self.assertEqual(len(self.hits.reported), 22)
        self.assertFalse(self.hits[0].dropped)
        self.assertFalse(self.hits[0].included)
        self.assertTrue(self.hits[0].reported)

    def test_manual_duplicate(self):
        self.assertEqual(len(self.hits.included), 15)
        self.assertEqual(len(self.hits.reported), 22)
        self.assertTrue(self.hits[0].included)
        self.assertTrue(self.hits[0].reported)

        self.hits[0].duplicate = True
        self.assertEqual(len(self.hits.included), 14)
        self.assertEqual(len(self.hits.reported), 21)
        self.assertFalse(self.hits[0].included)
        self.assertFalse(self.hits[0].reported)
        self.assertTrue(self.hits[0].duplicate)

        self.hits[0].duplicate = False
        self.assertEqual(len(self.hits.included), 14)
        self.assertEqual(len(self.hits.reported), 21)
        self.assertFalse(self.hits[0].included)
        self.assertFalse(self.hits[0].reported)
        self.assertFalse(self.hits[0].duplicate)
